import numpy as np
import pandas as pd

from ..base_structures import UnstructuredData


class LineSet:
    """PointSet with line cells.

    This dataset defines cell connectivity between points to create
    line segments.

    Args:
        data (UnstructuredData): Base object for unstructured data.

         data.cells represent the indices of the end points for each
         line segment in the mesh. Each column corresponds to a line
         segment. If not specified, the vertices are connected in order,
         equivalent to ``segments=[[0, 1], [1, 2], [2, 3], ...]``

        radius (float): Thickness of the line set
    """

    def __init__(self, data: UnstructuredData, radius: float = 1):

        self.data: UnstructuredData = data
        self.radius = radius

        if data.cells is None or data.cells.shape[1] < 2:
            self.generate_default_cells()

        elif data.cells.shape[1] != 2:
            raise AttributeError('data.cells must be of the format'
                                 'NDArray[(Any, 2), IntX]')

        # TODO: these must all be integer dtypes!

    def get_first_index_per_well(self, attr_hole_id: str):
        """ Method to get the first index of each well in the LineSet

        Returns:
            np.ndarray[(Any,), IntX]

        """
        import xarray as xr
        dataset: xr.Dataset = self.data.data
        dataframe: pd.DataFrame = self.data.attributes
        array = dataframe[attr_hole_id].values
        
        first_index = np.where(array[:-1] != array[1:])[0]
        return first_index
        
    
    def generate_default_cells(self):
        """ Method to generate cells based on the order of the vertex. This
        only works if the LineSet only represents one single line

        Returns:
            np.ndarray[(Any, 2), IntX]

        """
        a = np.arange(0, self.data.n_points - 1, dtype=np.int_)
        b = np.arange(1, self.data.n_points, dtype=np.int_)
        return np.vstack([a, b]).T

    @property
    def segments(self):
        return self.data.cells

    @property
    def n_segments(self):
        return self.segments.shape[0]



import pandas as pd
from typing import Union

import numpy as np
import warnings

from .surface_reader import read_mesh_file_to_vertex, read_mesh_file_to_cells, cells_from_delaunay, read_mesh_file_to_attr
from ....core.reader_helpers.reader_unstruct import ReaderUnstructuredHelper
from ....core.structs import UnstructuredData

from ....core.structs.base_structures.base_structures_enum import SpecialCellCase


def read_2d_mesh_to_unstruct(
        reader_args: ReaderUnstructuredHelper,
        delaunay: bool = True
) -> UnstructuredData:
    
    vertex: np.ndarray = read_mesh_file_to_vertex(reader_args.reader_vertex_args)
    cells: Union[np.ndarray, SpecialCellCase]
    cells_attr: Union[pd.DataFrame, None] = None
    vertex_attr: Union[pd.DataFrame, None] = None
    if reader_args.reader_cells_args is not None:
        cells = read_mesh_file_to_cells(reader_args.reader_cells_args)
    elif delaunay:
        cells = cells_from_delaunay(vertex)
    else:
        warnings.warn("No arguments to compute cell")
        cells = SpecialCellCase.POINTS
    if reader_args.reader_cells_attr_args is not None:
        cells_attr: pd.DataFrame = read_mesh_file_to_attr(reader_args.reader_cells_attr_args)
    if reader_args.reader_vertex_attr_args is not None:
        vertex_attr = read_mesh_file_to_attr(reader_args.reader_vertex_attr_args)

    ud = UnstructuredData.from_array(
        vertex=vertex,
        cells=cells,
        cells_attr=cells_attr,
        vertex_attr=vertex_attr,
    )
    return ud



def borehole_location_to_unstruct(reader_helper: ReaderFilesHelper,
                                  add_number_segments: bool = True) -> UnstructuredData:
    from . import _wells_api
    return _wells_api.borehole_location_to_unstruct(reader_helper, add_number_segments)


def read_survey_df_from_las(reader_helper: ReaderFilesHelper, well_name: str) -> 'pd.DataFrame':
    from .DEP import _well_files_reader
    return _well_files_reader.read_survey_df_from_las(reader_helper, well_name)


def read_assay_df_from_las(reader_helper: ReaderFilesHelper, well_name: str) -> 'pd.DataFrame':
    from .DEP import _well_files_reader
    return _well_files_reader.read_assay_df_from_las(reader_helper, well_name)


def welly_to_subsurface(wts: WellyToSubsurfaceHelper,
                        elev=True,
                        n_vertex_per_well=50,
                        convert_lith=True,
                        table: list['striplog.Component'] = None,
                        **kwargs) -> UnstructuredData:
    """Method to convert well data to `subsurface.UnstructuredData`

    Args:
        elev (bool): In general the (x, y, z) array of positions will have
            z as TVD, which is positive down. If `elev` is True, positive
            will be upwards.
        n_vertex_per_well (int): Number of vertex used to describe the geometry of the
         well.
        return_element (bool): if True return a `subsurface.LineSet` instead
        convert_lith (bool): if True convert lith from stiplog to curve
        table (List[Striplog.Component]): List of components to map lithologies
         to value.
        **kwargs:
            `Well.location.trajectory` kwargs

    Returns:

    """
    from . import _welly_reader
    return _welly_reader.welly_to_subsurface(wts, elev, n_vertex_per_well, convert_lith, table, **kwargs)




from dataclasses import dataclass
from typing import Optional, List
import numpy as np


__all__ = ['RexMesh', 'RexMaterial', 'RexLineSet']


@dataclass
class RexMesh:
    """Data structure that can be converted to rex binary. Represent
     a Datatype Mesh"""
    name: str
    vertex: np.ndarray
    edges: np.ndarray
    normals: Optional[np.ndarray] = np.array([])
    texture: Optional[np.ndarray] = np.array([])
    color: Optional[np.ndarray] = np.array([])
    material_id: Optional[int] = np.array([])

    @property
    def ver_ravel(self):
        return self.vertex.ravel().astype('float32')

    @property
    def tri_ravel(self):
        return self.edges.ravel().astype('int32')

    @property
    def color_ravel(self):
        return self.color.ravel()

    @property
    def n_vtx(self):
        return self.ver_ravel.shape[0]

    @property
    def n_triangles(self):
        return self.tri_ravel.shape[0]

    @property
    def n_color(self):
        return self.color_ravel.shape[0]


@dataclass
class RexMaterial:
    # ambient
    ka_red: float = 1
    ka_green: float = 1
    ka_blue: float = 1
    # specular
    ks_red: float = 1
    ks_green: float = 1
    ks_blue: float = 1

    # difuse
    kd_red: float = 1
    kd_green: float = 1
    kd_blue: float = 1

    # textures
    ka_texture_ID: int = 9223372036854775807
    ks_texture_ID: int = 9223372036854775807
    kd_texture_ID: int = 9223372036854775807

    ns: float = 0.1  # specular exponent
    alpha: float = 1


@dataclass
class RexLineSet:
    foo: int




"""
TODO: This is legacy code waiting to be updated to the new ideas


"""

import pandas as pd


__all__ = ['read_faultsticks_kingdom', 'read_faultsticks_charisma']


def read_faultsticks_kingdom(fp: str, name=None):
    """
    Reads in Kingdom fault stick files (kingdom) exported from Petrel (tested
    with Petrel 2017) and returns pandas DataFrame.

    Args:
        fp (str): Filepath.
        name (str, optional): Default: None.

    Returns:
        (pandas.DataFrame) Fault stick information stored in dataframe with
        ["X", "Y", "Z", "formation", "stick id"] columns.

    """
    storage = []
    with open(fp, "r") as file:
        lines = file.readlines()
        for line in lines:
            line = line.split(" ")
            X = float(line[6])
            Y = float(line[7])
            Z = float(line[9])
            if name is None:
                name = line[10]
            else:
                name = name
            stick = int(line[-2])
            storage.append([X, Y, Z, name, stick])

    df = pd.DataFrame(storage)
    df.columns = ["X", "Y", "Z", "name", "stick id"]
    return df


def read_faultsticks_charisma(fp:str, name=None):
    """
    Reads in charisma fault stick files exported from Petrel (tested with
    Petrel 2017) and returns pandas DataFrame.

    Args:
        fp (str): Filepath.
        name (str, optional): Default: None.

    Returns:
        (pandas.DataFrame) Fault stick information stored in dataframe with
        ["X", "Y", "Z", "formation", "stick id"] columns.

    """
    storage = []
    with open(fp, "r") as file:  # due to the variable delimiter length its
                                 # easier to just manually read this in
        lines = file.readlines()
        for line in lines:
            line = line.split(" ")
            line = [l for l in line if len(l) >= 1]
            X = float(line[3])
            Y = float(line[4])
            Z = float(line[5])
            if name is None:
                name = line[6]
            else:
                name = name
            stick = int(line[-1])
            storage.append([X, Y, Z, name, stick])

    df = pd.DataFrame(storage)
    df.columns = ["X", "Y", "Z", "name", "stick id"]
    return df



import numpy as np


__all__ = ['hex_to_rgb', ]


def hex_to_rgb(hex: str, normalize: bool = True) -> np.ndarray:
    """Transform colors from hex to rgb"""
    hex = hex.lstrip('#')
    hlen = len(hex)
    rgb = np.array(
        [int(hex[i:i + hlen // 3], 16) for i in range(0, hlen, hlen // 3)])
    if normalize is True:
        rgb = rgb / 255
    return rgb



from dataclasses import dataclass

from ..base_structures import UnstructuredData


@dataclass
class PointSet:
    """Class for pointset based data structures.

    This class uses UnstructuredData.vertex as cloud of points and the
    associated attributes. UnstructuredData.cells are not used.

    Args:
        data (UnstructuredData): Base object for unstructured data.

    """
    data: UnstructuredData
    
    def __init__(self, data: UnstructuredData):
        if data.cells.shape[1] > 1:
            raise AttributeError('data.cells must be of the format'
                                 'NDArray[(Any, 0), IntX] or NDArray[(Any, 1), IntX]')

        self.data = data

    @property
    def points(self) -> "np.ndarray":
        """Fetch the points/vertices dataframe."""
        return self.data.vertex

    @property
    def n_points(self):
        return self.data.vertex.shape[0]

    @property
    def point_data(self):
        """Fetch the scalar data associated with the vertices."""
        return self.data.attributes

    @property
    def point_data_dict(self):
        """Fetch the point data as a dictionary of numpy arrays."""
        return self.data.attributes_to_dict



from ..base_structures import UnstructuredData, StructuredData


class TriSurf:
    """PointSet with triangle cells.

    This dataset defines cell/element connectivity between points to create
    triangulated surface.

    Uses UnstructuredData.cells for the face connectivity.

    Args:
        mesh (UnstructuredData): Base object for unstructured data.
         data.cells  represent the point indices for each triangle
         in the mesh. Each column corresponds to a triangle edge.
        texture (StructuredData): 2D StructuredData with data to be mapped
         on the mesh

    Keyword Args:
        texture_origin : tuple(float)
            Length 3 iterable of floats defining the XYZ coordinates of the
            BOTTOM LEFT CORNER of the plane

        texture_point_u : tuple(float)
            Length 3 iterable of floats defining the XYZ coordinates of the
            BOTTOM RIGHT CORNER of the plane

        texture_point_v : tuple(float)
            Length 3 iterable of floats defining the XYZ coordinates of the
            TOP LEFT CORNER of the plane
    """

    def __init__(self,
                 mesh: UnstructuredData,
                 texture: StructuredData = None,
                 **kwargs
                 ):
        if mesh.cells.shape[1] != 3:
            raise AttributeError('data.cells must be of the format'
                                 'NDArray[(Any, 3), IntX]')

        self.mesh: UnstructuredData = mesh
        self.texture: StructuredData = texture
        self.texture_origin = kwargs.get('texture_origin', None)
        self.texture_point_u = kwargs.get('texture_point_u', None)
        self.texture_point_v = kwargs.get('texture_point_v', None)

    @property
    def has_texture_data_without_uv(self):
        return self.texture is not None and self.texture_origin is not None and self.texture_point_u is not None and self.texture_point_v is not None
    
    @property
    def has_texture_data_with_uv(self):
        return 'u' in self.mesh.points_attributes and 'v' in self.mesh.points_attributes
    
    @property
    def triangles(self):
        return self.mesh.cells

    @property
    def n_triangles(self):
        return self.mesh.cells.shape[0]



from subsurface.writer.to_rex.common import write_data_block_header, encode
from subsurface.writer.to_rex.data_struct import RexMaterial


__all__ = ['material_encode', 'write_material_data']


def material_encode(rex_material: RexMaterial, data_id: int):
    # Write data block header for Material
    data_header_bytes = write_data_block_header(
        data_type=5,  # Material data type
        version_data=1,  # Version. Probably useful for operation counter
        size_data=68,  # Size of the block is FIXED
        data_id=data_id  # self.data_id
    )

    # Write Material
    material_bytes = write_material_data(rex_material)
    rex_bytes = data_header_bytes + material_bytes
    return rex_bytes


def write_material_data(rex_material: RexMaterial):
    """Writes a standard material definition block

    Returns: bytes (size:68) representation of the material

    """
    input_ = [(rex_material.ka_red, 'float32'),
              (rex_material.ka_green, 'float32'),
              (rex_material.ka_blue, 'float32'),
              (rex_material.ka_texture_ID, 'uint64'),
              (rex_material.ks_red, 'float32'),
              (rex_material.ks_green, 'float32'),
              (rex_material.ks_blue, 'float32'),
              (rex_material.ks_texture_ID, 'uint64'),
              (rex_material.kd_red, 'float32'),
              (rex_material.kd_green, 'float32'),
              (rex_material.kd_blue, 'float32'),
              (rex_material.kd_texture_ID, 'uint64'),
              (rex_material.ns, 'float32'),
              (rex_material.alpha, 'float32')]
    block_bytes = encode(input_)
    return block_bytes



import re
import warnings
from typing import Optional, TextIO
import numpy as np
from subsurface.core.structs import UnstructuredData
from ._GOCAD_mesh import GOCADMesh


def mx_to_unstruc_from_binary(stream: TextIO) -> UnstructuredData:
    content = stream.read()
    goCADMeshes = _parse_lines(content)
    unstruct_data = _meshes_to_unstruct(goCADMeshes)
    return unstruct_data


def mx_to_unstruct_from_file(filename: str) -> UnstructuredData:
    # Split the file into meshes
    with open(filename, 'r') as f:
        content: str = f.read()
    goCAD_meshes = _parse_lines(content)
    unstruct_data = _meshes_to_unstruct(goCAD_meshes)

    return unstruct_data


def _parse_lines(content: str) -> list[GOCADMesh]:
    meshes = []
    mesh_blocks = re.split(r'(?=GOCAD TSurf 1)', content)
    # Remove any empty strings from the list

    mesh_blocks = [block for block in mesh_blocks if block.strip()]
    # Use multiprocessing Pool to parse meshes in parallel
    # with Pool(processes=cpu_count()) as pool:
    #     meshes = pool.map(process_mesh, mesh_blocks)
    for mesh_block in mesh_blocks:
        mesh_lines = mesh_block.split('\n')
        mesh = _process_mesh(mesh_lines)
        if mesh:
            meshes.append(mesh)
    return meshes


def _meshes_to_unstruct(meshes: list[GOCADMesh]) -> UnstructuredData:
    # ? I added this function to the Solutions class
    n_meshes = len(meshes)

    vertex_array = np.concatenate([meshes[i].vertices for i in range(n_meshes)])
    simplex_array = np.concatenate([meshes[i].vectorized_edges for i in range(n_meshes)])
    unc, count = np.unique(simplex_array, axis=0, return_counts=True)

    # * Prepare the simplex array
    simplex_array = meshes[0].vectorized_edges
    adder = 0
    for i in range(1, n_meshes):
        adder += np.max(meshes[i - 1].vectorized_edges) + 1
        add_mesh = meshes[i].vectorized_edges + adder
        simplex_array = np.append(simplex_array, add_mesh, axis=0)

    # * Prepare the cells_attr array
    ids_array = np.ones(simplex_array.shape[0])
    l0 = 0
    id = 1
    for mesh in meshes:
        l1 = l0 + mesh.vectorized_edges.shape[0]
        ids_array[l0:l1] = id
        l0 = l1
        id += 1

    # * Create the unstructured data
    import pandas as pd
    unstructured_data = UnstructuredData.from_array(
        vertex=vertex_array,
        cells=simplex_array,
        cells_attr=pd.DataFrame(ids_array, columns=['id'])  # TODO: We have to create an array with the shape of simplex array with the id of each simplex
    )

    return unstructured_data


def _process_mesh(mesh_lines) -> Optional[GOCADMesh]:
    mesh = GOCADMesh()
    in_header = False
    in_coord_sys = False
    in_property_class_header = False
    in_tface = False
    current_property_class_header = {}
    vertex_list = []
    vertex_indices = []
    triangle_list = []
    vid_to_index = {}  # Map from vertex id to index in vertex_list

    for line in mesh_lines:
        line = line.strip()

        if line.startswith('HEADER {'):
            in_header = True
            continue

        if in_header:
            if line == '}':
                in_header = False
            else:
                key_value = line.split(':', 1)
                if len(key_value) == 2:
                    key, value = key_value
                    mesh.header[key.strip()] = value.strip()
                else:
                    parts = line.split()
                    if len(parts) == 2:
                        key, value = parts
                        mesh.header[key.strip()] = value.strip()
                    else:
                        mesh.header[line.strip()] = None
            continue

        if line.startswith('GOCAD_ORIGINAL_COORDINATE_SYSTEM'):
            in_coord_sys = True
            continue
        if in_coord_sys:
            if line == 'END_ORIGINAL_COORDINATE_SYSTEM':
                in_coord_sys = False
            else:
                key_value = line.split(' ', 1)
                if len(key_value) == 2:
                    key, value = key_value
                    mesh.coordinate_system[key.strip()] = value.strip()
                else:
                    mesh.coordinate_system[line.strip()] = None
            continue

        if line.startswith('PROPERTY_CLASS_HEADER'):
            in_property_class_header = True
            current_property_class_header = {'name': line[len('PROPERTY_CLASS_HEADER'):].strip()}
            continue

        if in_property_class_header:
            if line == '}':
                in_property_class_header = False
                mesh.property_class_headers.append(current_property_class_header)
                current_property_class_header = {}
            else:
                key_value = line.split(':', 1)
                if len(key_value) == 2:
                    key, value = key_value
                    current_property_class_header[key.strip()] = value.strip()
                else:
                    key_value = line.split(' ', 1)
                    if len(key_value) == 2:
                        key, value = key_value
                        current_property_class_header[key.strip()] = value.strip()
                    else:
                        current_property_class_header[line.strip()] = None
            continue

        if line == 'TFACE':
            in_tface = True
            continue

        if in_tface:
            if line.startswith('VRTX') or line.startswith('PVRTX'):
                # Parse vertex line
                parts = line.split()
                if len(parts) >= 5:
                    _, vid, x, y, z = parts[:5]
                    vid = int(vid)
                    x, y, z = float(x), float(y), float(z)
                    vertex_indices.append(vid)
                    vertex_list.append([x, y, z])
                    vid_to_index[vid] = len(vertex_list) - 1
                    # If PVRTX then there could be more columns with property values. For now, we are just parsing the vertex coordinates.
                continue
            elif line.startswith('ATOM'):
                # Parse ATOM line
                parts = line.split()
                if len(parts) == 3:
                    _, vid, ref_vid = parts
                    vid = int(vid)
                    ref_vid = int(ref_vid)
                    if ref_vid in vid_to_index:
                        ref_index = vid_to_index[ref_vid]
                        coord = vertex_list[ref_index]
                        vertex_indices.append(vid)
                        vertex_list.append(coord)
                        vid_to_index[vid] = len(vertex_list) - 1
                    else:
                        warnings.warn(f"Reference vertex {ref_vid} not found for ATOM {vid}")
                continue
            elif line.startswith('TRGL'):
                # Parse triangle line
                parts = line.split()
                if len(parts) == 4:
                    _, v1, v2, v3 = parts
                elif len(parts) == 3:
                    v1, v2, v3 = parts
                else:
                    continue
                triangle_list.append([int(v1), int(v2), int(v3)])
                continue
            elif line.startswith('BSTONE'):
                _, value = line.split()
                mesh.bstones.append(int(value))
                continue
            elif line.startswith('BORDER'):
                parts = line.split()
                if len(parts) >= 4:
                    _, bid, v1, v2 = parts[:4]
                    mesh.borders.append({'id': int(bid), 'v1': int(v1), 'v2': int(v2)})
                continue
            elif line == 'END':
                in_tface = False
                continue
            else:
                pass
            continue

        # Other lines, possibly store as metadata
        if line:
            key_value = line.split(':', 1)
            if len(key_value) == 2:
                key, value = key_value
                mesh.metadata[key.strip()] = value.strip()
            else:
                pass

    # Convert lists to NumPy arrays
    mesh.vertices = np.array(vertex_list)
    mesh.vertex_indices = np.array(vertex_indices)
    mesh.edges = np.array(triangle_list)

    # Check the number of vertices, indices, and triangles to avoid errors (optional)

    return mesh



import subsurface.modules.reader
import subsurface.api.interfaces
import subsurface.modules.writer
from . import core
from .modules import visualization
from subsurface.core.structs import *
from datetime import datetime
import dotenv

dotenv.load_dotenv()

try:
    from subsurface import visualization
except ImportError:
    pass
try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:  # For Python <3.8, fallback
    from importlib_metadata import version, PackageNotFoundError
    
# Version.
try:
    __version__ = version("subsurface")  # Use package name
except ImportError:
    # If it was not installed, then we don't know the version. We could throw a
    # warning here, but this case *should* be rare. subsurface should be
    # installed properly!
    __version__ = 'unknown-'+datetime.today().strftime('%Y%m%d')

if __name__ == '__main__':
    pass



import pathlib
import warnings
from enum import Flag
from typing import TextIO, Union

import numpy as np

from .... import optional_requirements


class DXFEntityType(Flag):
    """Decides which entity types should be extracted."""
    FACE3D = 2 ** 0  # Extract only 3DFACE
    POLYLINE = 2 ** 1  # Extract only POLYLINE
    ALL = FACE3D | POLYLINE  # Extract all entity types


def dxf_from_file_to_vertex(
        file_path: Union[str, pathlib.Path],
        entity_type: DXFEntityType = DXFEntityType.ALL,
) -> np.ndarray:
    """
    Extract vertices from a DXF file, according to the chosen entity_type.
    
    :param file_path: Path to the DXF file.
    :param entity_type: Controls which entity types to extract.
    :return: Unique vertex array [N, 3].
    """
    ezdxf = optional_requirements.require_ezdxf()
    dataset = ezdxf.readfile(file_path)
    return _extract_vertices_from_dataset(dataset, entity_type)


def dxf_from_stream_to_vertex(
        stream: TextIO,
        entity_type: DXFEntityType = DXFEntityType.ALL,
) -> np.ndarray:
    """
    Extract vertices from a DXF stream, according to the chosen entity_type.
    
    :param stream: A file-like object containing the DXF data.
    :param entity_type: Controls which entity types to extract.
    :return: Unique vertex array [N, 3].
    """
    ezdxf = optional_requirements.require_ezdxf()
    dataset = ezdxf.read(stream)
    return _extract_vertices_from_dataset(dataset, entity_type)


def dxf_file_to_unstruct_input(
        file: Union[str, pathlib.Path],
) -> tuple[np.ndarray, np.ndarray, np.ndarray, dict]:
    """
    Extract unstructured-mesh-like data (vertex, cells, cell_attr_int, cell_attr_map) 
    from a DXF file using only 3DFACE entities.
    
    :param file: Path to the DXF file.
    :return: (vertex, cells, cell_attr_int, cell_attr_map)
    """
    ezdxf = optional_requirements.require_ezdxf()
    dataset = ezdxf.readfile(file)
    vertex, cells, cell_attr_int, cell_attr_map = _dxf_dataset_to_unstruct_input(dataset)

    if vertex.size == 0:
        raise ValueError("The DXF file does not contain any 3DFACE entities.")

    return vertex, cells, cell_attr_int, cell_attr_map


def dxf_stream_to_unstruct_input(
        stream: TextIO,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, dict]:
    """
    Extract unstructured-mesh-like data (vertex, cells, cell_attr_int, cell_attr_map) 
    from a DXF stream using only 3DFACE entities.
    
    :param stream: A file-like object containing the DXF data.
    :return: (vertex, cells, cell_attr_int, cell_attr_map)
    """
    ezdxf = optional_requirements.require_ezdxf()
    dataset = ezdxf.read(stream)
    return _dxf_dataset_to_unstruct_input(dataset)


def _extract_vertices_from_dataset(
        dataset,
        entity_type: DXFEntityType = DXFEntityType.ALL
) -> np.ndarray:
    """
    Collect unique vertices from the dataset's modelspace.
    The entity_type flag dictates which entities are extracted.
    """
    vertices = []

    for e in dataset.modelspace():
        match e.dxftype():
            # 3DFACE
            case "3DFACE" if DXFEntityType.FACE3D in entity_type:
                # A 3DFACE entity typically has three corners: e[0], e[1], e[2]
                # Sometimes it can have a fourth corner e[3], which might be equal to e[2].
                # Adjust accordingly if needed:
                vertices.extend([e[0], e[1], e[2]])
                # If you'd like to handle the potential 4th corner,
                # you could do: vertices.append(e[3]) if e[3] != e[2] else None

            # POLYLINE
            case "POLYLINE" if DXFEntityType.POLYLINE in entity_type:
                for v in e.vertices:
                    x, y, z = v.dxf.location.xyz
                    vertices.append([x, y, z])

            # Other / unsupported
            case _:
                # If it doesn't match the chosen entity type(s), we skip or warn.
                # But if you'd prefer to only warn when an entity isn't recognized at all,
                # you can do so here:
                warnings.warn(f"Entity type '{e.dxftype()}' not extracted (flag: {entity_type}).")
                continue

    # Convert to numpy array and ensure uniqueness
    vertices = np.array(vertices)
    if vertices.size == 0:
        return vertices
    return np.unique(vertices, axis=0)


def _dxf_dataset_to_unstruct_input(dataset: 'ezdxf.drawing.Drawing') -> tuple[np.ndarray, np.ndarray, np.ndarray, dict]:
    """
    Build unstructured-mesh-like data from 3DFACE entities in a dataset:
      - vertex coordinates
      - connectivity in 'cells' array
      - cell attributes in both integer-coded and mapping (string->int) forms

    """
    """
    Build unstructured-mesh-like data from 3DFACE entities in a dataset:
      - vertex coordinates
      - connectivity in 'cells' array
      - cell attributes in both integer-coded and mapping (string->int) forms

    Returns (cell_attr_int, cell_attr_map, cells, vertex).
    """

    def _map_cell_attr_strings_to_integers(cell_attr):
        """
        Map string layer names (or other string attributes) to integer IDs, starting from 1.
        """
        unique_values = np.unique(cell_attr)
        sorted_unique_values = sorted(unique_values)

        # Create mapping from string values to integers (start at 1, 2, 3, ...)
        value_to_int_mapping = {
                str(value): index + 1 for index, value in enumerate(sorted_unique_values)
        }

        cell_attr_int = np.array([value_to_int_mapping[value] for value in cell_attr])
        return cell_attr_int, value_to_int_mapping

    vertex_list = []
    cell_attr = []

    for e in dataset.modelspace():
        match e.dxftype():
            case "3DFACE":
                # Typically 3 corners, but can have 4th corner repeated
                vertex_list.extend([e[0], e[1], e[2]])
                cell_attr.append(e.dxf.get("layer"))
            case _:
                # We ignore other entities in unstructured input
                continue

    vertices = np.array(vertex_list)
    # For each triple (3DFACE), build a cell (triangle)
    cells = np.arange(0, vertices.shape[0]).reshape(-1, 3)

    cell_attr_int, cell_attr_map = _map_cell_attr_strings_to_integers(cell_attr)
    return vertices, cells, cell_attr_int, cell_attr_map



from typing import Union, Callable

import numpy as np
import pandas as pd


def mesh_csv_to_vertex(path_to_file: str, columns_map: Union[None, Callable, dict, pd.Series] = None,
                       **reader_kwargs) -> np.ndarray:
    data = pd.read_csv(path_to_file, **reader_kwargs)
    if columns_map is not None: map_columns_names(columns_map, data)
    return get_vertices_from_df(data)


def mesh_csv_to_cells(path_to_file: str, columns_map: Union[None, Callable, dict, pd.Series] = None,
                      **reader_kwargs) -> np.ndarray:
    data = pd.read_csv(path_to_file, **reader_kwargs)
    if columns_map is not None: map_columns_names(columns_map, data)
    return get_cells_from_df(data)


def mesh_csv_to_attributes(path_to_file: str,
                           columns_map: Union[None, Callable, dict, pd.Series] = None,
                           **reader_kwargs) -> pd.DataFrame:
    data = pd.read_csv(path_to_file, **reader_kwargs)
    if columns_map is not None:
        map_columns_names(columns_map, data)
    return data


def get_cells_from_df(data):
    try:
        cells = data[['e1', 'e2', 'e3']].dropna().astype('int').values
    except KeyError:
        raise KeyError('Columns e1, e2, and e3 must be present in the data set. Use'
                       'columns_map to map other names')
    return cells


def get_vertices_from_df(data):
    try:
        vertex = data[['x', 'y', 'z']].values
    except KeyError:
        raise KeyError('Columns x, y, and z must be present in the data set. Use'
                       'columns_map to map other names')
    return vertex


def map_columns_names(columns_map: Union[Callable, dict, pd.Series], data: pd.DataFrame):
    data.columns = data.columns.map(columns_map)
    if data.columns.isin(['x', 'y', 'z']).any() is False:
        raise AttributeError('At least x, y, z must be passed to `columns_map`')

    return data.columns



from dataclasses import dataclass
from typing import Union, Dict, Mapping, Hashable, Any, Literal

import numpy as np
import pandas as pd
import xarray as xr

from subsurface.core.structs.base_structures._unstructured_data_constructor import vertex_and_cells_arrays_to_data_array, raw_attributes_to_dict_data_arrays
from subsurface.core.structs.base_structures.base_structures_enum import SpecialCellCase


@dataclass(frozen=False)
class UnstructuredData:
    data: xr.Dataset
    cells_attr_name: str = "cell_attrs"
    vertex_attr_name: str = "vertex_attrs"

    """Primary structure definition for unstructured data

    Attributes:
        data (`xarray.Dataset`): Data structure where we store

    Args:

        ds (xarray.Dataset): Directly a dataset with the expected structured. This
         arg is specially thought for loading data from disk

    Notes:
        Depending on the shape of `edge` the following unstructured elements can
        be created:

        - cells NDArray[(Any, 0), IntX] or NDArray[(Any, 1), IntX] -> *Point cloud*.
          E.g. Outcrop scan with lidar
        - cells NDArray[(Any, 2), IntX] -> *Lines*. E.g. Borehole
        - cells NDArray[(Any, 3), IntX] -> *Mesh*. E.g surface-DEM Topography
        - cells NDArray[(Any, 4), IntX]
           - -> *tetrahedron*
           - -> *quadrilateral (or tetragon)* UNSUPPORTED?
        - cells NDArray[(Any, 8), IntX] -> *Hexahedron: Unstructured grid/Prisms*
    """

    def __post_init__(self):
        self._validate()

    def __repr__(self):
        return self.data.__repr__()

    @classmethod
    def from_array(
            cls,
            vertex: np.ndarray,
            cells: Union[np.ndarray, Literal["lines", "points"], SpecialCellCase],
            *,
            cells_attr: Union[None, pd.DataFrame, Dict[str, xr.DataArray]] = None,
            vertex_attr: Union[None, pd.DataFrame, Dict[str, xr.DataArray]] = None,
            coords: Mapping[Hashable, Any] = None,
            xarray_attributes: Mapping[Hashable, Any] = None,
            default_cells_attr_name: str = "cell_attrs",
            default_points_attr_name: str = "vertex_attrs",
            attributes: Union[None, pd.DataFrame, Dict[str, xr.DataArray]] = None  # TODO Obsolete
    ):
        """ Constructor of UnstructuredData from arrays or pandas DataFrames.

        Args:
            vertex (np.ndarray): NDArray[(Any, 3), FloatX]: XYZ point data
            cells (Union[np.ndarray, Literal["lines", "points"]]): NDArray[(Any, ...), IntX]:
             Combination of vertex that create different geometric elements. If
             str use default values for either points or lines
            cells_attr (Union[None, pd.DataFrame, Dict[str, xr.DataArray]]: Number associated to an element
            vertex_attr (Union[None, pd.DataFrame, Dict[str, xr.DataArray]]: Number
             associated to points
            coords:
            xarray_attributes:
            attributes:
            default_cells_attr_name:
            default_points_attr_name:

        Returns:

        """
        if attributes is not None:
            cells_attr = attributes

        cells_data_array, n_cells, n_vertex, vertex_data_array = vertex_and_cells_arrays_to_data_array(
            cells=cells,
            vertex=vertex
        )
        points_attributes_xarray_dict: dict[str, xr.DataArray] = raw_attributes_to_dict_data_arrays(
            default_attributes_name=default_points_attr_name,
            n_items=n_vertex,
            dims=["points", "vertex_attr"],
            raw_attributes=vertex_attr
        )

        cells_attributes_xarray_dict: dict[str, xr.DataArray] = raw_attributes_to_dict_data_arrays(
            default_attributes_name=default_cells_attr_name,
            n_items=n_cells,
            dims=["cell", "cell_attr"],
            raw_attributes=cells_attr
        )

        xarray_dict = {
                "vertex": vertex_data_array,
                "cells" : cells_data_array,
                **cells_attributes_xarray_dict,
                **points_attributes_xarray_dict
        }

        return cls.from_data_arrays_dict(
            xarray_dict=xarray_dict,
            coords=coords,
            xarray_attributes=xarray_attributes,
            default_cells_attributes_name=default_cells_attr_name,
            default_points_attributes_name=default_points_attr_name
        )

    @classmethod
    def from_data_arrays_dict(
            cls,
            xarray_dict: Dict[str, xr.DataArray],
            coords: Mapping[Hashable, Any] = None,
            xarray_attributes: Mapping[Hashable, Any] = None,
            default_cells_attributes_name="cell_attrs",
            default_points_attributes_name="vertex_attrs"
    ):
        # TODO: xr.Dataset seems to have been changed with 2022.06. needs to be adapted for indexing
        ds = xr.Dataset(xarray_dict, coords=coords, attrs=xarray_attributes)

        # Try to unstack pandas dataframe if exist
        # TODO: This is an issue in wells. If it is only there maybe we should move it there
        try:
            ds = ds.reset_index('cell')
        except (KeyError, ValueError) as e:
            print(f"{e} xarray dataset must include 'cell' key (KeyError) or xarray 'cell' has no index (ValueError).")

        # Check that the Dataset data_vars names matches, "vertex", "cells", default_cells_attributes_name and default_points_attributes_name
        # and raise an error pointing out which one is missing
        _vars = [var in ds.data_vars for var in ["vertex", "cells", default_cells_attributes_name, default_points_attributes_name]]
        if not all(_vars):
            missing = ["vertex", "cells", default_cells_attributes_name, default_points_attributes_name]
            raise KeyError(f"DataArray must include the following keys: {missing}. For attributes you can"
                           f" change the default names with default_cells_attributes_name and default_points_attributes_name.")

        return cls(ds, default_cells_attributes_name, default_points_attributes_name)

    @classmethod
    def from_binary_le(cls, path: str):
        from ._liquid_earth_mesh import LiquidEarthMesh
        with open(path, 'rb') as f:
            bytes_data = f.read()
        mesh = LiquidEarthMesh.from_binary(bytes_data)
        unstruct = cls.from_array(
            vertex=mesh.vertex,
            cells=mesh.cells,
            cells_attr=mesh.attributes,
            vertex_attr=mesh.points_attributes,
            xarray_attributes=None
        )
        return unstruct

    @classmethod
    def from_binary_le_legacy(cls, path_to_binary: str, path_to_json: str):
        import json
        from ._liquid_earth_mesh import LiquidEarthMesh

        with open(path_to_binary, 'rb') as f:
            body_ = f.read()
        with open(path_to_json, 'r') as f:
            header_ = json.load(f)

        header_json = json.dumps(header_)
        header_json_bytes = header_json.encode('utf-8')
        header_json_length = len(header_json_bytes)
        header_json_length_bytes = header_json_length.to_bytes(4, byteorder='little')
        file = header_json_length_bytes + header_json_bytes + body_
        mesh = LiquidEarthMesh.from_binary(file)
        unstruct = cls.from_array(
            vertex=mesh.vertex,
            cells=mesh.cells,
            cells_attr=mesh.attributes,
            vertex_attr=mesh.points_attributes,
            xarray_attributes=None
        )
        return unstruct

    @property
    def vertex(self) -> np.ndarray:
        return self.data['vertex'].values

    @property
    def cells(self):
        return self.data['cells'].values

    @property
    def attributes(self) -> pd.DataFrame:
        xarray = self.data[self.cells_attr_name]
        return xarray.to_dataframe()[self.cells_attr_name].unstack(level=1)

    @attributes.setter
    def attributes(self, dataframe):
        self.data[self.cells_attr_name] = xr.DataArray(dataframe, dims=['element', 'cell_attr'])

    @property
    def cell_attributes(self):
        return self.attributes

    @cell_attributes.setter
    def cell_attributes(self, dataframe):
        self.attributes = dataframe

    @property
    def points_attributes(self) -> pd.DataFrame:
        data_array: xr.DataArray = self.data[self.vertex_attr_name]
        dataframe: pd.DataFrame = data_array.to_dataframe()
        not_suer = dataframe[self.vertex_attr_name]
        unstack = not_suer.unstack(level=1)
        return unstack

    @points_attributes.setter
    def points_attributes(self, dataframe: pd.DataFrame):
        vertex_attr: xr.DataArray = self.data[self.vertex_attr_name]
        vertex_attr.values = dataframe.values

    @property
    def n_elements(self):
        return self.cells.shape[0]

    @property
    def n_vertex_per_element(self):
        return self.cells.shape[1]

    @property
    def n_points(self):
        return self.vertex.shape[0]

    @property
    def attributes_to_dict(
            self,
            orient: Literal["dict", "list", "series", "split", "tight", "index"] = "list"
    ):
        return self.attributes.to_dict(orient)

    @property
    def points_attributes_to_dict(
            self,
            orient: Literal["dict", "list", "series", "split", "tight", "index"] = "list"
    ):
        return self.points_attributes.to_dict(orient)

    @property
    def extent(self):
        max = self.vertex.max(axis=0)
        min = self.vertex.min(axis=0)
        extent = np.stack((min, max), axis=1).ravel()
        return extent

    def to_xarray(self):
        a = xr.DataArray(self.vertex, dims=['points', 'XYZ'])
        b = xr.DataArray(self.cells, dims=['cells', 'node'])
        e = xr.DataArray(self.attributes, dims=['element', 'cell_attr'])
        c = xr.Dataset({'v': a, 'e': b, 'a': e})
        return c

    def to_binary_legacy(self, order='F'):
        bytearray_le = self._to_bytearray(order)
        header = self._set_binary_header()
        return bytearray_le, header

    def to_binary(self, order='F') -> bytes:
        body_ = self._to_bytearray(order)
        header_ = self._set_binary_header()
        import json
        header_json = json.dumps(header_)
        header_json_bytes = header_json.encode('utf-8')
        header_json_length = len(header_json_bytes)
        header_json_length_bytes = header_json_length.to_bytes(4, byteorder='little')
        file = header_json_length_bytes + header_json_bytes + body_
        return file

    def _set_binary_header(self):
        header = {
                "vertex_shape"     : self.vertex.shape,
                "cell_shape"       : self.cells.shape,
                "cell_attr_shape"  : self.attributes.shape,
                "vertex_attr_shape": self.points_attributes.shape,
                "cell_attr_names"  : self.attributes.columns.to_list(),
                "cell_attr_types"  : self.attributes.dtypes.astype(str).to_list(),
                "vertex_attr_names": self.points_attributes.columns.to_list(),
                "vertex_attr_types": self.attributes.dtypes.astype(str).to_list(),
                "xarray_attrs"     : self.data.attrs
        }
        return header

    def _to_bytearray(self, order):
        vertex = self.vertex.astype('float32').tobytes(order)
        cells = self.cells.astype('int32').tobytes(order)
        cell_attribute = self.attributes.values.astype('float32').tobytes(order)
        vertex_attribute = self.points_attributes.values.astype('float32').tobytes(order)
        bytearray_le = vertex + cells + cell_attribute + vertex_attribute
        return bytearray_le

    def _validate(self):
        try:
            _ = self.data[self.cells_attr_name]['cell']
            _ = self.data[self.cells_attr_name]['cell_attr']
        except KeyError:
            raise KeyError('Cell attribute DataArrays must contain dimension cell and cell_attr')
        try:
            _ = self.data[self.vertex_attr_name]['vertex_attr']
            _ = self.data[self.vertex_attr_name]['points']
        except KeyError:
            raise KeyError('Point attribute DataArrays must contain dimensions points and vertex_attr.')

        # Make sure the number of vertices matches the associated data.
        if self.data['cells']['cell'].size != self.data[self.cells_attr_name]['cell'].size:
            raise AttributeError('Attributes and cells must have the same length.')

        if self.n_points != self.data[self.vertex_attr_name]['points'].size:
            raise AttributeError('points_attributes and vertex must have the same length.')



from typing import Sequence, Optional

import numpy as np

from ....core.structs.structured_elements.structured_grid import StructuredGrid
from ....optional_requirements import require_rasterio
from ....core.structs import StructuredData, UnstructuredData
from ....core.utils.utils_core import get_extension
from ....core.reader_helpers.readers_data import GenericReaderFilesHelper
from ....core.reader_helpers.reader_unstruct import ReaderUnstructuredHelper
from ..mesh.surfaces_api import read_2d_mesh_to_unstruct


def read_structured_topography(path, crop_to_extent: Optional[Sequence]=None) -> StructuredData:
    rasterio = require_rasterio()

    extension = get_extension(path)
    if extension == '.tif':
        structured_data = rasterio_dataset_to_structured_data(
            dataset=rasterio.open(path),
            crop_to_extent=crop_to_extent
        )
    else:
        raise NotImplementedError('The extension given cannot be read yet')

    return structured_data


def read_structured_topography_to_unstructured(path) -> UnstructuredData:
    structured_data = read_structured_topography(path)
    return topography_to_unstructured_data(structured_data)


def rasterio_dataset_to_structured_data(dataset, crop_to_extent: Optional[Sequence]=None):
    if crop_to_extent is not None: 
        window = _get_raster_window(crop_to_extent, dataset)
    else:
        window = None

    data = dataset.read(1, window=window)
    data = np.fliplr(data.T)
    shape = data.shape
    
    # TODO: ===================
    # TODO: Add the option to crop
    # TODO: Resample
    
    coords = {
        'x': np.linspace(
            dataset.bounds.left,
            dataset.bounds.right,
            shape[0]
        ),
        'y': np.linspace(
            dataset.bounds.bottom,
            dataset.bounds.top,
            shape[1]
        )
    }
    structured_data = StructuredData.from_numpy(data, data_array_name='topography', coords=coords)
    return structured_data


def read_unstructured_topography(path, additional_reader_kwargs: Optional[dict]=None) -> UnstructuredData:
    """For example, a dxf file"""
    
    additional_reader_kwargs = additional_reader_kwargs or {}
    helper = GenericReaderFilesHelper(
        file_or_buffer=path,
        additional_reader_kwargs=additional_reader_kwargs or {}
    )
    unstructured_helper = ReaderUnstructuredHelper(helper)
    unstruct: UnstructuredData = read_2d_mesh_to_unstruct(unstructured_helper)
    return unstruct


def topography_to_unstructured_data(structured_data: StructuredData) -> UnstructuredData:
    from subsurface.modules.visualization import to_pyvista_grid

    sg = StructuredGrid(structured_data)
    s = to_pyvista_grid(sg, data_order='C', data_set_name='topography')
    un_s = s.cast_to_unstructured_grid()
    un_s.triangulate(inplace=True)
    vertex = un_s.points
    cells = un_s.cells.reshape(-1, 4)[:, 1:]

    unstructured_data = UnstructuredData.from_array(vertex, cells)
    return unstructured_data


def _get_raster_window(crop_to_extent, dataset):
    from rasterio.windows import Window
    # TODO: Add None check
    # Get the indices of the window
    left, bottom, right, top = crop_to_extent
    row_start, col_start = dataset.index(left, top)
    row_stop, col_stop = dataset.index(right, bottom)
    # Read the data in the window
    window = Window.from_slices((row_start, row_stop), (col_start, col_stop))
    return window



from subsurface.core.geological_formats.boreholes.boreholes import MergeOptions

from subsurface.core.structs.base_structures.base_structures_enum import SpecialCellCase

from subsurface.core.structs import UnstructuredData, PointSet

from subsurface import optional_requirements
from subsurface.modules.reader.wells.read_borehole_interface import read_collar, read_survey, read_lith, read_attributes

from subsurface.core.geological_formats import BoreholeSet, Collars, Survey

from subsurface.core.reader_helpers.readers_data import GenericReaderFilesHelper


def read_wells(
        collars_reader: GenericReaderFilesHelper,
        surveys_reader: GenericReaderFilesHelper,
        attrs_reader: GenericReaderFilesHelper,
        is_lith_attr: bool,
        number_nodes: int = 10,
        add_attrs_as_nodes: bool = False,
        duplicate_attr_depths: bool = False 
) -> BoreholeSet:
    # ! FIGUROUT IF WE NEED LITH
    
    
    pd = optional_requirements.require_pandas()
    try:
        collar_df: pd.DataFrame = read_collar(collars_reader)
        unstruc: UnstructuredData = UnstructuredData.from_array(
            vertex=collar_df[["x", "y", "z"]].values,
            cells=SpecialCellCase.POINTS
        )
        points = PointSet(data=unstruc)
        collars: Collars = Collars(
            ids=collar_df.index.to_list(),
            collar_loc=points
        )
    except KeyError as e:
        raise ValueError(f"Error while reading collars: {e}. Check if the columns are named x, y, z or use Columns Map.")
    except Exception as e:
        raise ValueError(f"Error while reading collars: {e}")   
    
    try:
        survey_df: pd.DataFrame = read_survey(surveys_reader)
    except Exception as e:
        raise ValueError(f"Error while reading surveys: {e}")
    
    try:
        attrs: pd.DataFrame = read_attributes(attrs_reader, is_lith=is_lith_attr)
    except Exception as e:
        raise ValueError(f"Error while reading attributes: {e}")
    
    try:
        if add_attrs_as_nodes:
            attr_df = attrs
        else:
            attr_df = None

        survey: Survey = Survey.from_df(
            survey_df=survey_df,
            attr_df=attr_df,
            number_nodes=number_nodes,
            duplicate_attr_depths=duplicate_attr_depths
        )

        # Check if component lith is in columns or columns_map
        survey.update_survey_with_lith(attrs)
    except Exception as e:
        raise ValueError(f"Error while creating survey: {e}")

    borehole_set = BoreholeSet(
        collars=collars,
        survey=survey,
        merge_option=MergeOptions.INTERSECT
    )

    return borehole_set



import os
from io import BytesIO
from typing import Union

from subsurface.core.structs import StructuredData

from .... import optional_requirements
from ....core.structs import UnstructuredData
from subsurface.core.reader_helpers.readers_data import GenericReaderFilesHelper
import numpy as np
import pandas as pd


def read_VTK_structured_grid(file_or_buffer: Union[str, BytesIO], active_scalars: str) -> StructuredData:
    pv = optional_requirements.require_pyvista()

    if isinstance(file_or_buffer, BytesIO):
        # If file_or_buffer is a BytesIO, write it to a temporary file
        from tempfile import NamedTemporaryFile
        with NamedTemporaryFile('wb', suffix='.vtk', delete=False) as temp_file:
            # Write the BytesIO content to the temporary file
            getvalue: bytes = file_or_buffer.getvalue()
            temp_file.write(getvalue)
            temp_file.flush()  # Make sure all data is written
            temp_file_name = temp_file.name  # Store the temporary file name
        try:
            # Use pyvista.read() to read from the temporary file
            pyvista_obj = pv.read(temp_file_name)
        finally:
            # Ensure the temporary file is deleted after reading
            os.remove(temp_file_name)
    else:
        # If it's a file path, read directly
        pyvista_obj = pv.read(file_or_buffer)
    try:
        pyvista_struct: pv.ExplicitStructuredGrid = pv_cast_to_explicit_structured_grid(pyvista_obj)
    except Exception as e:
        raise f"The file is not a structured grid: {e}"

    if PLOT := False:
        pyvista_struct.set_active_scalars(active_scalars)
        pyvista_struct.plot()

    struct: StructuredData = StructuredData.from_pyvista_structured_grid(
        grid=pyvista_struct,
        data_array_name=active_scalars
    )

    return struct


def read_volumetric_mesh_to_subsurface(reader_helper_coord: GenericReaderFilesHelper,
                                       reader_helper_attr: GenericReaderFilesHelper) -> UnstructuredData:
    df_coord = read_volumetric_mesh_coord_file(reader_helper_coord)
    if len(df_coord.columns) == 1:
        raise ValueError(
            "The attributes file has only one column, probably the columns are not being separated correctly. Use 'sep' in Additional Reader Arguments"
        )

    df_attr = read_volumetric_mesh_attr_file(reader_helper_attr)
    # Check if there are more than one column and if it is only one raise an error that probably the columns have not been properly separated. Use "sep" in Additional Reader Arguments
    if len(df_attr.columns) == 1:
        raise ValueError(
            "The attributes file has only one column, probably the columns are not being separated correctly. Use 'sep' in Additional Reader Arguments"
        )
    
    combined_df = df_coord.merge(df_attr, left_index=True, right_index=True)
    ud = UnstructuredData.from_array(
        vertex=combined_df[['x', 'y', 'z']], cells="points",
        attributes=combined_df[['pres', 'temp', 'sg', 'xco2']]
    )
    return ud


def read_volumetric_mesh_coord_file(reader_helper: GenericReaderFilesHelper) -> pd.DataFrame:
    df = pd.read_csv(
        filepath_or_buffer=reader_helper.file_or_buffer,
        **reader_helper.pandas_reader_kwargs
    )
    if reader_helper.columns_map is not None:
        df.rename(
            mapper=reader_helper.columns_map,
            axis="columns",
            inplace=True
        )

    df.dropna(axis=0, inplace=True)

    df.x = df.x.astype(float)
    df.y = df.y.astype(float)
    df.z = df.z.astype(float)
    # Throw error if empty
    if df.empty:
        raise ValueError("The file is empty")
    
    return df


def read_volumetric_mesh_attr_file(reader_helper: GenericReaderFilesHelper) -> pd.DataFrame:
    df = pd.read_table(reader_helper.file_or_buffer, **reader_helper.pandas_reader_kwargs)
    df.columns = df.columns.astype(str).str.strip()
    return df


def pv_cast_to_explicit_structured_grid(pyvista_object: 'pv.DataSet') -> 'pv.ExplicitStructuredGrid':
    pv = optional_requirements.require_pyvista()

    match pyvista_object:
        case pv.RectilinearGrid() as rectl_grid:
            return __pv_convert_rectilinear_to_explicit(rectl_grid)
        case pv.UnstructuredGrid() as unstr_grid:
            return __pv_convert_unstructured_to_explicit(unstr_grid)
        case _:
            return pyvista_object.cast_to_explicit_structured_grid()


def __pv_convert_unstructured_to_explicit(unstr_grid):
    """
    Convert a PyVista UnstructuredGrid to an ExplicitStructuredGrid if possible.
    """
    pv = optional_requirements.require_pyvista()

    # First check if the grid has the necessary attributes to be treated as structured
    if not hasattr(unstr_grid, 'n_cells') or unstr_grid.n_cells == 0:
        raise ValueError("The unstructured grid has no cells.")
    
    # Try to detect if the grid has a structured topology
    # Check if the grid has cell type 11 (VTK_VOXEL) or 12 (VTK_HEXAHEDRON)
    cell_types = unstr_grid.celltypes
    
    # Voxels (11) and hexahedra (12) are the cell types used in structured grids
    if not all(ct in [11, 12] for ct in cell_types):
        raise ValueError("The unstructured grid contains non-hexahedral cells and cannot be converted to explicit structured.")
    
    # Try to infer dimensions from the grid
    try:
        # Method 1: Try PyVista's built-in conversion if available
        return unstr_grid.cast_to_explicit_structured_grid()
    except (AttributeError, TypeError):
        pass
    
    try:
        # Method 2: If the grid has dimensions stored as field data
        if "dimensions" in unstr_grid.field_data:
            dims = unstr_grid.field_data["dimensions"]
            if len(dims) == 3:
                nx, ny, nz = dims
                # Verify that dimensions match the number of cells
                if (nx-1)*(ny-1)*(nz-1) != unstr_grid.n_cells:
                    raise ValueError("Stored dimensions do not match the number of cells.")
                
                # Extract points and reorder if needed
                points = unstr_grid.points.reshape((nx, ny, nz, 3))
                
                # Create explicit structured grid
                explicit_grid = pv.ExplicitStructuredGrid((nx, ny, nz), points.reshape((-1, 3)))
                explicit_grid.compute_connectivity()
                
                # Transfer data arrays
                for name, array in unstr_grid.cell_data.items():
                    explicit_grid.cell_data[name] = array.copy()
                for name, array in unstr_grid.point_data.items():
                    explicit_grid.point_data[name] = array.copy()
                for name, array in unstr_grid.field_data.items():
                    if name != "dimensions":  # Skip dimensions field
                        explicit_grid.field_data[name] = array.copy()
                
                return explicit_grid
    except (ValueError, KeyError):
        pass
    
    # If none of the above methods work, use PyVista's extract_cells function
    # to reconstruct the structured grid if possible
    try:
        # This is a best-effort approach that tries multiple strategies
        return pv.core.filters.convert_unstructured_to_structured_grid(unstr_grid)
    except Exception as e:
        raise ValueError(f"Failed to convert unstructured grid to explicit structured grid: {e}")
    
def __pv_convert_rectilinear_to_explicit(rectl_grid):

    pv = optional_requirements.require_pyvista()

    # Extract the coordinate arrays from the input RectilinearGrid.
    x = np.asarray(rectl_grid.x)
    y = np.asarray(rectl_grid.y)
    z = np.asarray(rectl_grid.z)

    # Helper function: "double" the coordinates to produce an expanded set
    # that, when processed internally via np.unique, returns the original nodal values.
    def doubled_coords(arr):
        return np.repeat(arr, 2)[1:-1]

    # Double the coordinate arrays.
    xcorn = doubled_coords(x)
    ycorn = doubled_coords(y)
    zcorn = doubled_coords(z)

    # Build a complete grid of corner points via meshgrid. Fortran ('F') order ensures
    # the connectivity ordering aligns with VTK's expectations.
    xx, yy, zz = np.meshgrid(xcorn, ycorn, zcorn, indexing='ij')
    corners = np.column_stack((xx.ravel(order='F'),
                               yy.ravel(order='F'),
                               zz.ravel(order='F')))

    # The dimensions to pass to the ExplicitStructuredGrid constructor should be
    # the counts of unique coordinates in each direction.
    dims = (len(np.unique(xcorn)),
            len(np.unique(ycorn)),
            len(np.unique(zcorn)))

    # Create the ExplicitStructuredGrid.
    explicit_grid = pv.ExplicitStructuredGrid(dims, corners)
    explicit_grid.compute_connectivity()

    # --- Copy associated data arrays ---

    # Transfer all cell data arrays.
    for name, array in rectl_grid.cell_data.items():
        explicit_grid.cell_data[name] = array.copy()

    # Transfer all point data arrays.
    for name, array in rectl_grid.point_data.items():
        explicit_grid.point_data[name] = array.copy()

    # (Optional) Transfer field data as well.
    for name, array in rectl_grid.field_data.items():
        explicit_grid.field_data[name] = array.copy()

    return explicit_grid



from dataclasses import dataclass, field
from typing import List, Dict, Any, Tuple, Optional, Union, TextIO
import numpy as np
import xarray as xr
from pathlib import Path

from ....core.structs import StructuredData


@dataclass
class GridDimensions:
    """
    Represents the dimensions of a 3D grid.

    Attributes:
        nx (int): Number of cells in the x-direction
        ny (int): Number of cells in the y-direction
        nz (int): Number of cells in the z-direction
    """
    nx: int
    ny: int
    nz: int


@dataclass
class GridOrigin:
    """
    Represents the origin point of a 3D grid.

    Attributes:
        x (float): X-coordinate of the origin
        y (float): Y-coordinate of the origin
        z (float): Z-coordinate of the origin
    """
    x: float
    y: float
    z: float


@dataclass
class GridCellSizes:
    """
    Represents the cell sizes in each direction of a 3D grid.

    Attributes:
        x (List[float]): Cell sizes in the x-direction
        y (List[float]): Cell sizes in the y-direction
        z (List[float]): Cell sizes in the z-direction
    """
    x: List[float]
    y: List[float]
    z: List[float]


@dataclass
class GridData:
    """
    Represents a 3D grid with dimensions, origin, and cell sizes.

    Attributes:
        dimensions (GridDimensions): The dimensions of the grid
        origin (GridOrigin): The origin point of the grid
        cell_sizes (GridCellSizes): The cell sizes in each direction
        metadata (Dict[str, Any]): Optional metadata about the grid
    """
    dimensions: GridDimensions
    origin: GridOrigin
    cell_sizes: GridCellSizes
    metadata: Dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_dict(cls, grid_dict: Dict[str, Any]) -> 'GridData':
        """
        Converts a dictionary containing grid information into a GridData instance.

        Args:
            grid_dict: Dictionary with grid information

        Returns:
            GridData: A new GridData instance
        """
        dims = grid_dict["dimensions"]
        origin_dict = grid_dict["origin"]
        cell_sizes_dict = grid_dict["cell_sizes"]

        # Handle both new and legacy key names
        nx = dims.get("nx", dims.get("ne"))
        ny = dims.get("ny", dims.get("nn"))
        nz = dims.get("nz", dims.get("nz"))

        x = origin_dict.get("x", origin_dict.get("x0"))
        y = origin_dict.get("y", origin_dict.get("y0"))
        z = origin_dict.get("z", origin_dict.get("z0"))

        x_sizes = cell_sizes_dict.get("x", cell_sizes_dict.get("easting"))
        y_sizes = cell_sizes_dict.get("y", cell_sizes_dict.get("northing"))
        z_sizes = cell_sizes_dict.get("z", cell_sizes_dict.get("vertical"))

        metadata = grid_dict.get("metadata", {})

        return cls(
            dimensions=GridDimensions(nx=nx, ny=ny, nz=nz),
            origin=GridOrigin(x=x, y=y, z=z),
            cell_sizes=GridCellSizes(x=x_sizes, y=y_sizes, z=z_sizes),
            metadata=metadata
        )


def read_msh_structured_grid(grid_stream: TextIO, values_stream: TextIO, missing_value: Optional[float],
                             attr_name: Optional[str]) -> StructuredData:
    """
    Read a structured grid mesh and values from streams and return a StructuredData object.

    This function is designed to work with streams (e.g., from Azure blob storage)
    rather than file paths.

    Args:
        grid_stream: TextIO stream containing the grid definition (.msh format)
        values_stream: TextIO stream containing the property values (.mod format)

    Returns:
        StructuredData object containing the grid and property values

    Raises:
        ValueError: If the stream format is invalid
    """
    # Read all lines from the grid stream
    lines = [line.strip() for line in grid_stream if line.strip()]

    # Create metadata for the grid
    metadata = {
            'file_format': 'grav3d',
            'source'     : 'stream'
    }

    # Parse grid information from lines
    try:
        grid = _parse_grid_from_lines(lines, metadata)
    except ValueError as e:
        # Add context about the stream to the error message
        raise ValueError(f"Error parsing grid stream: {e}") from e

    # Read values from the values stream
    try:
        # Read all values from the stream
        lines = [line.strip() for line in values_stream if line.strip()]

        model_array = _parse_mod_file(grid, lines, missing_value=missing_value)

    except Exception as e:
        # Add context to any errors
        raise ValueError(f"Error reading model stream: {str(e)}") from e

    # Create and return a StructuredData object
    return structured_data_from(model_array, grid, data_name=attr_name)


def read_msh_file(filepath: Union[str, Path]) -> GridData:
    """
    Read a structured grid mesh file and return a GridData object.

    Currently supports Grav3D mesh file format (.msh):
    - First line: NX NY NZ (number of cells in X, Y, Z directions)
    - Second line: X Y Z (coordinates of origin in meters)
    - Next section: X cell widths (either expanded or using N*value notation)
    - Next section: Y cell widths (either expanded or using N*value notation)
    - Next section: Z cell thicknesses (either expanded or using N*value notation)

    Args:
        filepath: Path to the mesh file

    Returns:
        GridData object containing the mesh information

    Raises:
        FileNotFoundError: If the file doesn't exist
        ValueError: If the file format is invalid
    """
    filepath = Path(filepath)
    if not filepath.exists():
        raise FileNotFoundError(f"Mesh file not found: {filepath}")

    with open(filepath, 'r') as f:
        lines = [line.strip() for line in f.readlines() if line.strip()]

    metadata = {
            'file_format': 'grav3d',
            'filepath'   : str(filepath)
    }

    try:
        return _parse_grid_from_lines(lines, metadata)
    except ValueError as e:
        # Add context about the file to the error message
        raise ValueError(f"Error parsing mesh file {filepath}: {e}") from e


def read_mod_file(filepath: Union[str, Path], grid: GridData,
                  missing_value: float = -99_999.0) -> np.ndarray:
    """
    Read a model file containing property values for a 3D grid.

    Currently supports Grav3D model file format (.mod) where each line contains
    a single property value. The values are ordered with the z-direction changing
    fastest, then x, then y.

    Args:
        filepath: Path to the model file
        grid: GridData object containing the grid dimensions
        missing_value: Value to replace with NaN in the output array (default: -99_999.0)

    Returns:
        3D numpy array of property values with shape (ny, nx, nz)

    Raises:
        FileNotFoundError: If the file doesn't exist
        ValueError: If the number of values doesn't match the grid dimensions
    """
    filepath = Path(filepath)
    if not filepath.exists():
        raise FileNotFoundError(f"Model file not found: {filepath}")

    try:
        # Read all values from the file
        with open(filepath, 'r') as f:
            lines = [line.strip() for line in f if line.strip()]

        model_array = _parse_mod_file(grid, lines, missing_value)

        return model_array

    except Exception as e:
        # Add context to any errors
        raise ValueError(f"Error reading model file {filepath}: {str(e)}") from e


def _parse_mod_file(grid: GridData, lines: List[str], missing_value: Optional[float]) -> np.ndarray:
    # Convert each line to a float
    values = np.array([float(line) for line in lines], dtype=float)
    # Calculate expected number of values based on grid dimensions
    nx, ny, nz = grid.dimensions.nx, grid.dimensions.ny, grid.dimensions.nz
    expected_count = nx * ny * nz
    if len(values) != expected_count:
        raise ValueError(
            f"Invalid model file: expected {expected_count} values, got {len(values)}"
        )
    # Reshape to (ny, nx, nz) with z changing fastest
    model_array = values.reshape((ny, nx, nz))
    # Replace missing values with NaN
    if missing_value is not None:
        model_array[model_array == missing_value] = np.nan
    return model_array


def structured_data_from(array: np.ndarray, grid: GridData,
                         data_name: str = 'model') -> StructuredData:
    """
    Convert a 3D numpy array and grid information into a StructuredData object.

    Args:
        array: 3D numpy array of property values with shape (ny, nx, nz)
        grid: GridData object containing grid dimensions, origin, and cell sizes
        data_name: Name for the data array (default: 'model')

    Returns:
        StructuredData object containing the data array with proper coordinates

    Raises:
        ValueError: If array shape doesn't match grid dimensions
    """
    # Verify array shape matches grid dimensions
    expected_shape = (grid.dimensions.ny, grid.dimensions.nx, grid.dimensions.nz)
    if array.shape != expected_shape:
        raise ValueError(
            f"Array shape {array.shape} doesn't match grid dimensions {expected_shape}"
        )

    # Calculate cell center coordinates
    centers = _calculate_cell_centers(grid)

    # Create the xarray DataArray with proper coordinates
    xr_data_array = xr.DataArray(
        data=array,
        dims=['y', 'x', 'z'],  # Dimensions in the order they appear in the array
        coords={
                'x': centers['x'],
                'y': centers['y'],
                'z': centers['z'],
        },
        name=data_name,
        attrs=grid.metadata  # Include grid metadata in the data array
    )

    # Create a StructuredData instance from the xarray DataArray
    struct = StructuredData.from_data_array(
        data_array=xr_data_array,
        data_array_name=data_name
    )

    return struct


def _parse_grid_from_lines(lines: List[str], metadata: Dict[str, Any] = None) -> GridData:
    """
    Parse grid information from a list of lines.

    Args:
        lines: List of lines containing grid information
        metadata: Optional metadata to include in the GridData object

    Returns:
        GridData object containing the parsed grid information

    Raises:
        ValueError: If the lines format is invalid
    """
    if len(lines) < 2:
        raise ValueError("Invalid format: insufficient data")

    # Parse dimensions (first line)
    try:
        dims = lines[0].split()
        nx, ny, nz = int(dims[0]), int(dims[1]), int(dims[2])
    except (IndexError, ValueError) as e:
        raise ValueError(f"Invalid dimensions: {e}")

    # Parse origin coordinates (second line)
    try:
        origin = lines[1].split()
        x, y, z = float(origin[0]), float(origin[1]), float(origin[2])
    except (IndexError, ValueError) as e:
        raise ValueError(f"Invalid origin: {e}")

    # Parse cell sizes
    try:
        current_line = 2
        x_sizes, current_line = _parse_cell_sizes(lines, current_line, nx)
        y_sizes, current_line = _parse_cell_sizes(lines, current_line, ny)
        z_sizes, _ = _parse_cell_sizes(lines, current_line, nz)
    except (IndexError, ValueError) as e:
        raise ValueError(f"Error parsing cell sizes: {e}")

    # Create a GridData object with the parsed information
    grid_data_dict = {
            'dimensions': {'nx': nx, 'ny': ny, 'nz': nz},
            'origin'    : {'x': x, 'y': y, 'z': z},
            'cell_sizes': {
                    'x': x_sizes,
                    'y': y_sizes,
                    'z': z_sizes
            },
            'metadata'  : metadata or {}
    }

    return GridData.from_dict(grid_data_dict)


def _parse_cell_sizes(lines: List[str], start_index: int, count: int) -> Tuple[List[float], int]:
    """
    Parse cell sizes from file lines, handling both compact (N*value) and expanded notation.

    Args:
        lines: List of lines from the file
        start_index: Index to start parsing from
        count: Number of values to parse

    Returns:
        Tuple containing:
            - List of parsed values
            - Next line index after parsing
    """
    line = lines[start_index]

    # Check for compact notation (N*value)
    if '*' in line:
        parts = line.split('*')
        repetition = int(parts[0])
        value = float(parts[1])
        values = [value] * repetition
        return values, start_index + 1

    # Handle expanded notation across multiple lines
    values = []
    line_index = start_index

    while len(values) < count and line_index < len(lines):
        current_line = lines[line_index]

        # If we encounter a line with compact notation while parsing expanded, 
        # it's likely the next section
        if '*' in current_line and len(values) > 0:
            break

        # Add all numbers from the current line
        values.extend([float(x) for x in current_line.split()])
        line_index += 1

    # Take only the required number of values
    return values[:count], line_index


def _calculate_cell_centers(grid: GridData) -> Dict[str, np.ndarray]:
    """
    Calculate the center coordinates of each cell in the grid.

    Args:
        grid: GridData object containing grid dimensions, origin, and cell sizes

    Returns:
        Dictionary with 'x', 'y', and 'z' keys containing arrays of cell center coordinates
    """
    # Convert cell sizes to numpy arrays for vectorized operations
    x_sizes = np.array(grid.cell_sizes.x)
    y_sizes = np.array(grid.cell_sizes.y)
    z_sizes = np.array(grid.cell_sizes.z)

    # Calculate cell centers by adding cumulative sizes and offsetting by half the first cell size
    x_centers = grid.origin.x + np.cumsum(x_sizes) - x_sizes[0] / 2
    y_centers = grid.origin.y + np.cumsum(y_sizes) - y_sizes[0] / 2

    # For z, cells typically extend downward from the origin
    z_centers = grid.origin.z - (np.cumsum(z_sizes) - z_sizes[0] / 2)

    return {
            'x': x_centers,
            'y': y_centers,
            'z': z_centers
    }



import enum
from dataclasses import dataclass
from typing import Dict, List, Tuple, Union, Literal

import numpy as np
import xarray as xr

from ....optional_requirements import require_pyvista


class StructuredDataType(enum.Enum):
    REGULAR_AXIS_ALIGNED = 0  #: Regular axis aligned grid. Distance between consecutive points is constant
    REGULAR_AXIS_UNALIGNED = 1  #: Regular axis unaligned grid. Distance between consecutive points is constant
    IRREGULAR_AXIS_ALIGNED = 2  #: Irregular axis aligned grid. Distance between consecutive points is not constant
    IRREGULAR_AXIS_UNALIGNED = 3  #: Irregular axis unaligned grid. Distance between consecutive points is not constant


@dataclass(frozen=False)
class StructuredData:
    data: xr.Dataset
    _active_data_array_name: str = "data_array"
    type: StructuredDataType = StructuredDataType.REGULAR_AXIS_ALIGNED
    dtype: Literal["float32", "float64"] = "float32"

    """Primary structure definition for structured data

       Check out other constructors: `StructuredData.from_numpy`,
        `StructuredData.from_data_array` and `StructuredData.from_dict`

    Args:
        data (xr.Dataset): object containing
         structured data, i.e. data that can be stored in multidimensional
         numpy array. The preferred type to pass as data is directly a
         xr.Dataset to be sure all the attributes are set and named as the user
         wants.
        data_array_name (str): If data is a numpy array or xarray DataArray, data_name
         provides the name for the xarray data variable
     
    Attributes:
        data (xarray.Dataset)
    """

    @property
    def active_data_array_name(self):
        data_var_list = list(self.data.data_vars.keys())
        if self._active_data_array_name not in data_var_list:
            raise ValueError("data_array_name not found in data_vars: {}".format(data_var_list))
        return self._active_data_array_name

    @active_data_array_name.setter
    def active_data_array_name(self, data_array_name: str):
        self._active_data_array_name = data_array_name

    @classmethod
    def from_numpy(cls, array: np.ndarray, coords: dict = None, data_array_name: str = "data_array",
                   dim_names: List[str] = None):
        if dim_names is None:
            dim_names = cls._default_dim_names(array.ndim)
        # if they are more than 3 we do not know the dimension name but it should valid:

        dataset: xr.Dataset = xr.Dataset(
            data_vars=
            {
                    data_array_name: (dim_names, array)
            },
            coords=coords
        )

        return cls(dataset, data_array_name)

    @classmethod
    def from_data_array(cls, data_array: xr.DataArray, data_array_name: str = "data_array"):
        dataset: xr.Dataset = xr.Dataset(
            data_vars={
                    data_array_name: data_array
            },
            coords=data_array.coords
        )

        return cls(dataset, data_array_name)

    @classmethod
    def from_dict(cls, data_dict: Dict[str, xr.DataArray], coords: Dict[str, str] = None, data_array_name: str = "data_array"):
        dataset: xr.Dataset = xr.Dataset(data_vars=data_dict, coords=coords)
        return cls(dataset, data_array_name)

    @classmethod
    def from_pyvista_structured_grid(
            cls,
            grid: Union["pyvista.ExplicitStructuredGrid", "pyvista.StructuredGrid"],
            data_array_name: str = "data_array"
    ):
        pyvista = require_pyvista()
        # Extract p

        # Extract cell data and point data (if any)
        data_vars = {}

        # TODO: I need to do something with the bounds

        dimensions = np.array(grid.dimensions) - 1
        default_dim_names = cls._default_dim_names(dimensions.shape[0])

        bounds: tuple = grid.bounds
        coords = {}
        for i, dim in enumerate(default_dim_names):
            coords[dim] = np.linspace(
                start=bounds[i * 2],
                stop=bounds[i * 2 + 1],
                num=dimensions[i],
                endpoint=False
            )

        for name in grid.cell_data:
            cell_attr_data: pyvista.pyvista_ndarray = grid[name]
            cell_attr_data_reshaped = cell_attr_data.reshape(dimensions, order='F')

            data_vars[name] = xr.DataArray(
                data=cell_attr_data_reshaped,
                dims=default_dim_names,
                name=name
            )

        dataset: xr.Dataset = xr.Dataset(
            data_vars=data_vars,
            coords=coords
        )
        return cls(dataset, data_array_name)

    @property
    def values(self) -> np.ndarray:
        return self.data[self.active_data_array_name].values

    _bounds: Tuple[float, float, float, float, float, float] = None

    @property
    def bounds(self):
        if self._bounds is not None:
            return self._bounds

        array_: xr.DataArray = self.data[self.active_data_array_name]
        bounds = self._get_bounds_from_coord(array_)
        return bounds

    @bounds.setter
    def bounds(self, bounds: Tuple[float, float, float, float, float, float]):
        """
        Set the bounds of the structured data. This is useful for defining the
        spatial extent of the data in a structured grid.

        Args:
            bounds (Tuple[float, float, float, float, float, float]): A tuple containing
                the minimum and maximum values for each dimension (xmin, xmax, ymin, ymax, zmin, zmax).
        """
        self._bounds = bounds

    @property
    def shape(self):
        return self.active_data_array.shape

    @property
    def active_data_array(self):
        return self.data[self.active_data_array_name]

    @staticmethod
    def _get_bounds_from_coord(xr_obj: xr.DataArray):
        bounds = {}
        for coord in xr_obj.coords:
            bounds[coord] = (xr_obj[coord].min().item(), xr_obj[coord].max().item())
        return bounds

    def default_data_array_to_binary_legacy(self, order: Literal["K", "A", "C", "F"] = 'F'):
        bytearray_le = self._to_bytearray(order=order)
        header = self._set_binary_header()

        return bytearray_le, header

    def to_binary(self, order: Literal["K", "A", "C", "F"] = 'F') -> bytes:
        """Converts the structured data to a binary file
        
        Notes: 
            Only the active data array is converted to binary for now 
        """

        body_ = self._to_bytearray(order)
        header = self._set_binary_header()

        import json
        header_json = json.dumps(header)
        header_json_bytes = header_json.encode('utf-8')
        header_json_length = len(header_json_bytes)
        header_json_length_bytes = header_json_length.to_bytes(4, byteorder='little')
        file = header_json_length_bytes + header_json_bytes + body_
        return file

    def _set_binary_header(self) -> Dict:
        data_array = self.active_data_array

        match self.type:
            case StructuredDataType.REGULAR_AXIS_ALIGNED:
                header = {
                        "data_shape": self.shape,
                        "bounds"    : self.bounds,
                        "transform" : None,
                        "dtype"     : self.dtype,
                        "data_name" : self.active_data_array_name
                }
            case _:
                raise NotImplementedError(f"StructuredDataType {self.type} not implemented yet")

        return header

    def _to_bytearray(self, order: Literal["K", "A", "C", "F"]) -> bytes:
        data_array = self.active_data_array

        data = data_array.values.astype(self.dtype).tobytes(order)
        bytearray_le = data
        return bytearray_le

    @classmethod
    def _default_dim_names(cls, n_dims: int):
        if n_dims == 2:
            dim_names = ['x', 'y']
        elif n_dims == 3:
            dim_names = ['x', 'y', 'z']
        else:
            dim_names = ['dim' + str(i) for i in range(n_dims)]
        return dim_names

    def to_netcdf(self, path: str, **to_netcdf_kwargs):
        """
        Serializes the current StructuredData instance to a NetCDF file.

        Args:
            path (str): The path (including file name) where the NetCDF file will be saved.
            **to_netcdf_kwargs: Additional keyword arguments forwarded to xarray's `to_netcdf`.
        """
        # Copy the dataset (shallow copy of the data structure, no copying of the underlying arrays)
        ds = self.data.copy(deep=False)

        # Store relevant metadata as global attributes:
        ds.attrs["active_data_array_name"] = self._active_data_array_name
        ds.attrs["structured_data_type"] = self.type.name  # e.g., "REGULAR_AXIS_ALIGNED"
        ds.attrs["dtype"] = self.dtype  # e.g., "float32"

        # Use xarray's to_netcdf
        ds.to_netcdf(path, **to_netcdf_kwargs)

    @classmethod
    def from_netcdf(cls, path: str, **from_netcdf_kwargs):
        """
        Deserializes a NetCDF file into a StructuredData instance.

        Args:
            path (str): The path to the NetCDF file to read.
            **from_netcdf_kwargs: Additional keyword arguments forwarded to xarray's `open_dataset`.

        Returns:
            StructuredData: A new instance of StructuredData loaded from the file.
        """
        ds = xr.open_dataset(path, **from_netcdf_kwargs)

        # Retrieve what was stored in attrs (with defaults if missing)
        data_array_name = ds.attrs.get("active_data_array_name", "data_array")
        dtype_str: str = ds.attrs.get("dtype", "float32")
        if dtype_str not in ["float32", "float64"]:
            raise ValueError(f"Unsupported dtype: {dtype_str}")

        sdt_str = ds.attrs.get("structured_data_type", "REGULAR_AXIS_ALIGNED")

        # Convert strings back to your enum or any other type
        # (assuming StructuredDataType is an Enum where name matches sdt_str)
        if sdt_str not in StructuredDataType.__members__:
            raise ValueError(f"Unsupported structured_data_type: {sdt_str}")
        structured_data_type: StructuredDataType = StructuredDataType[sdt_str]

        return cls(
            data=ds,
            _active_data_array_name=data_array_name,
            type=structured_data_type,
            dtype=dtype_str
        )



import glob
from typing import List

import warnings

from pandas import DataFrame

from subsurface import optional_requirements
from subsurface.reader.wells.pandas_to_welly import WellyToSubsurfaceHelper
import numpy as np

from subsurface.structs.base_structures import UnstructuredData

welly = optional_requirements.require_welly()
striplog = optional_requirements.require_striplog()


def welly_to_subsurface(
        wts: WellyToSubsurfaceHelper,
        elev=True,
        n_vertex_per_well=50,
        convert_lith=True,
        table: List[striplog.Component] = None,
        **kwargs) -> UnstructuredData:
    """Method to convert well data to `subsurface.UnstructuredData`

    Args:
        elev (bool): In general the (x, y, z) array of positions will have
            z as TVD, which is positive down. If `elev` is True, positive
            will be upwards.
        n_vertex_per_well (int): Number of vertex used to describe the geometry of the
         well.
        return_element (bool): if True return a `subsurface.LineSet` instead
        convert_lith (bool): if True convert lith from stiplog to curve
        table (List[Striplog.Component]): List of components to map lithologies
         to value.
        **kwargs:
            `Well.location.trajectory` kwargs

    Returns:

    """
    vertex = np.zeros((0, 3))
    cells = np.zeros((0, 2), dtype=int)

    last_index = 0
    missed_wells = []

    for w in wts.p.get_wells():
        if well_without_valid_survey(w, missed_wells): continue

        cells, vertex, last_index = vertex_and_cells_from_welly_trajectory(
            cells, elev, kwargs, last_index, n_vertex_per_well, vertex, w)

        change_curve_basis_to_n_vertex_per_well_inplace(n_vertex_per_well, w, wts)

        # Convert striplog into Curve
        if convert_lith is True:
            if 'lith' in w.data:
                if table is None: table = wts.lith_component_table
                w.data['lith_log'] = striplog_to_curve_log(n_vertex_per_well, table, w, wts)
            else:
                w.data["lith_log"] = welly.Curve(-1 * np.ones(n_vertex_per_well - 1, dtype=int))

            if w.data["lith_log"].shape[0] != n_vertex_per_well - 1:
                raise ValueError("Cell_attr does not match cells")

    try:
        df = wts.p.df()
    except ValueError as e:
        if 'objects passed' in str(e):
            df = None
        else:
            raise ValueError

    print('The following boreholes failed being processed: ', missed_wells)

    unstructured_data = UnstructuredData.from_array(vertex, cells, df)
    return unstructured_data


def striplog_to_curve_log(n_vertex_per_well, table, w: welly.Well, wts: WellyToSubsurfaceHelper) -> welly.Curve:
    start, stop, step_size = wts._calculate_basis_parameters(w, n_vertex_per_well - 1)
    s_log, basis, _table = w.data['lith'].to_log(step_size, start, stop, table=table, return_meta=True)
    return welly.Curve(s_log, basis)


def change_curve_basis_to_n_vertex_per_well_inplace(n_points, w, wts):
    start, stop, step_size = wts._calculate_basis_parameters(w, n_points)
    basis = np.arange(start, stop, step_size)
    w.unify_basis(keys=None, basis=basis)


def vertex_and_cells_from_welly_trajectory(
        cells: np.ndarray, elev: bool,
        welly_trajectory_kwargs: dict,
        last_index: int, n_vertex_for_well: int,
        vertex: np.ndarray, w: welly.Well):
    try:
        datum = w.location.datum
    except AttributeError:
        datum = None

    xyz = w.location.trajectory(datum=datum, elev=elev, points=n_vertex_for_well, **welly_trajectory_kwargs)  # w.location.position
    # Make sure deviation is there
    a = np.arange(0 + last_index, xyz.shape[0] - 1 + last_index, dtype=np.int_)
    b = np.arange(1 + last_index, xyz.shape[0] + last_index, dtype=np.int_)
    cells_b = np.vstack([a, b]).T

    vertex = np.vstack((vertex, xyz))
    cells = np.vstack((cells, cells_b))
    last_index += xyz.shape[0]

    return cells, vertex, last_index


def well_without_valid_survey(w: welly.Well, missed_wells: List[str]):
    well_without_position = w.location.position is None
    if well_without_position:
        warnings.warn(f'At least one of the wells do not have '
                      'assigned a survey. Borehole name: {w.name}')
        missed_wells.append(w.name)
    return well_without_position


def _create_welly_well_from_las(well_name: str, las_folder: str):
    """
    Add well from welly las file.

    Args:
        well_name (str): Well name
        las_folder (str): Path to las file

    Returns:

    """

    def _create_well(uwi: str = 'dummy_uwi'):
        w = welly.Well()
        w.location.uwi = well_name
        w.location.name = well_name

        return w

    well = _create_well(well_name)
    return add_curves_from_las(well, las_folder)


def add_curves_from_las(w: welly.Well, las_folder: str) -> welly.Well:
    """ Add curves from las file. """

    def _read_curves_to_welly_object(well: welly.Well, curve_path: str = '.') -> welly.Well:
        las_files = glob.glob(curve_path + '*.las')
        # throw error if no las files found
        if len(las_files) == 0:
            raise Exception('No las files found in ' + curve_path)

        for curve in las_files:
            well.add_curves_from_las(curve)
        return well

    _read_curves_to_welly_object(w, curve_path=las_folder)

    w_df = w.df()
    deviation_df = _make_deviation_df(w_df, inclination_header='IMG_INCL', azimuth_header='IMG_AZ')
    w.location.add_deviation(deviation_df)
    return w


def _make_deviation_df(well_df: DataFrame, inclination_header: str, azimuth_header: str,
                       depth_header: str = '', depth_index=True):
    if depth_index:
        deviation_data = well_df[[inclination_header, azimuth_header]].to_numpy()
        deviation_index = well_df.index.to_numpy()
        deviation_complete = np.insert(deviation_data, 0, deviation_index, axis=1)
        deviation_complete = deviation_complete[~np.isnan(deviation_complete).any(axis=1), :]
    else:
        deviation_complete = well_df[[depth_header, inclination_header, azimuth_header]].to_numpy()
        deviation_complete = deviation_complete[~np.isnan(deviation_complete).any(axis=1), :]
    return deviation_complete



import io
import pathlib
from typing import Callable

from subsurface.core.reader_helpers.readers_data import GenericReaderFilesHelper, SupportedFormats
import pandas as pd


def check_format_and_read_to_df(reader_helper: GenericReaderFilesHelper) -> pd.DataFrame:
    # ? This swithch is veeery confusing
    match (reader_helper.file_or_buffer, reader_helper.format):
        case _, SupportedFormats.JSON | ".json":
            d = pd.read_json(reader_helper.file_or_buffer, orient='split')
        case str() | pathlib.Path(), _:
            reader: Callable = _get_reader(reader_helper.format)
            d = reader(
                filepath_or_buffer=reader_helper.file_or_buffer,
                sep=reader_helper.separator,
                **reader_helper.pandas_reader_kwargs
            )
        case (bytes() | io.BytesIO() | io.StringIO() | io.TextIOWrapper()), _:
            reader = _get_reader(reader_helper.format)
            d = reader(reader_helper.file_or_buffer, **reader_helper.pandas_reader_kwargs)
        case dict(), _:
            reader = _get_reader('dict')
            d = reader(reader_helper.file_or_buffer)
        case _:
            raise AttributeError('file_or_buffer must be either a path or a dict')
            
    if type(d.columns) is str:  d.columns = d.columns.str.strip()  # Remove spaces at the beginning and end
    if type(d.index) is str: d.index = d.index.str.strip()  # Remove spaces at the beginning and end
    return d




def _get_reader(file_format) -> Callable:
    def _dict_reader(dict_):
        return pd.DataFrame(
            data=dict_['data'],
            columns=dict_['columns'],
            index=dict_['index']
        )

    match file_format:
        case SupportedFormats.XLXS:
            raise NotImplemented("Pandas changed the backend for reading excel files and needs to be re-implemented")
            reader = pd.read_excel
        case 'dict':
            reader = _dict_reader
        case SupportedFormats.CSV:
            reader = pd.read_csv
        case SupportedFormats.JSON:
            reader = _dict_reader
        case _:
            raise ValueError(f"Subsurface is not able to read the following extension: {file_format}")
    return reader



# import warnings

from .profiles import *

from .topography.topo_core import read_structured_topography, read_unstructured_topography

from .mesh.omf_mesh_reader import omf_stream_to_unstructs
from .mesh.dxf_reader import dxf_stream_to_unstruct_input, dxf_file_to_unstruct_input
from .mesh.mx_reader import mx_to_unstruc_from_binary
from .mesh.obj_reader import load_obj_with_trimesh, load_obj_with_trimesh_from_binary
from .mesh.glb_reader import load_gltf_with_trimesh

from .volume.read_grav3d import read_msh_structured_grid



from typing import Union, Dict, Literal, List

import numpy as np
import pandas as pd
import xarray as xr

from subsurface.core.structs.base_structures.base_structures_enum import SpecialCellCase


def vertex_and_cells_arrays_to_data_array(cells: Union[np.ndarray, Literal["lines", "points"], SpecialCellCase],
                                          vertex: np.ndarray):
    n_vertex = vertex.shape[0]
    if type(cells) is not np.ndarray:
        cells: np.ndarray = _create_default_cells_arg(
            cells=cells,
            n_vertex=n_vertex
        )
    n_cells = cells.shape[0]

    vertex_data_array = xr.DataArray(
        data=vertex,
        dims=['points', 'XYZ'],
        coords={'XYZ': ['X', 'Y', 'Z']}
    )
    cells_data_array = xr.DataArray(cells, dims=['cell', 'nodes'])
    return cells_data_array, n_cells, n_vertex, vertex_data_array


def raw_attributes_to_dict_data_arrays(
        default_attributes_name: str, n_items: int, dims: List[str],
        raw_attributes: Union[None, pd.DataFrame, Dict[str, xr.DataArray]]) \
        -> Dict[str, xr.DataArray]:
    
    if raw_attributes is None or type(raw_attributes) is pd.DataFrame:
        points_attributes_xarray_dict = {
                default_attributes_name: _data_array_attributes_from_raw_data(
                    raw_data=raw_attributes,
                    dims=dims,
                    n_rows=n_items
                )
        }
    else:
        points_attributes_xarray_dict = raw_attributes
    return points_attributes_xarray_dict


def _create_default_cells_arg(cells: Union[Literal["points", "lines"], SpecialCellCase],
                              n_vertex: int) -> np.ndarray:
    if cells is None or cells == 'points' or cells == SpecialCellCase.POINTS:
        cells_array = np.arange(0, n_vertex).reshape(-1, 1)
    elif cells == 'lines' or cells == SpecialCellCase.LINES:
        a = np.arange(0, n_vertex - 1, dtype=np.int_)
        b = np.arange(1, n_vertex, dtype=np.int_)
        cells_array = np.vstack([a, b]).T
    else:
        raise ValueError("cells must be either None (will default to 'points'),"
                         "'points', 'lines' or a 2D ndarray.")
    return cells_array


def _data_array_attributes_from_raw_data(raw_data: Union[None, pd.DataFrame],
                                         dims: List[str], n_rows: int) -> xr.DataArray:
    if raw_data is None:
        raw_data = pd.DataFrame(np.zeros((n_rows, 0)))

    if type(raw_data) is pd.DataFrame:
        data_array = xr.DataArray(raw_data, dims=dims)
    else:
        raise ValueError("cells_attributes must be either pd.DataFrame or " "None/default.")
    return data_array



import enum
from typing import List
import numpy as np
from subsurface.optional_requirements import require_scipy

from ....core.structs import UnstructuredData, StructuredData

__all__ = ['interpolate_unstructured_data_to_structured_data', ]


class InterpolationMethod(enum.Enum):
    linear = "linear"
    nearest = "nearest"


def interpolate_unstructured_data_to_structured_data(
        ud: UnstructuredData, attr_name: str,
        resolution: List[int] = None,
        interpolation_method: InterpolationMethod = InterpolationMethod.nearest) -> StructuredData:
    if resolution is None:
        resolution = [50, 50, 50]
    boundaries_max = ud.vertex.max(axis=0)
    boundaries_min = ud.vertex.min(axis=0)
    coords = dict()
    dims = ['x', 'y', 'z']
    for e, i in enumerate(dims):
        coords[i] = np.linspace(boundaries_min[e], boundaries_max[e], resolution[e], endpoint=False)

    grid = np.meshgrid(*coords.values())
    scipy = require_scipy()
    interpolated_attributes = scipy.interpolate.griddata(
        points=ud.vertex,
        values=ud.attributes.loc[:, attr_name],
        xi=tuple(grid),
        method=interpolation_method.value
    )

    sd = StructuredData.from_numpy(
        array=interpolated_attributes,
        data_array_name=attr_name,
        coords=coords
    )
    return sd



import enum

import warnings

from typing import Union, Tuple, Optional

from ... import optional_requirements
from ...core.structs.unstructured_elements import PointSet, TriSurf, LineSet, TetraMesh
from ...core.structs.structured_elements.structured_grid import StructuredGrid
import numpy as np

try:
    import pyvista as pv
except ImportError:
    warnings.warn('Pyvista is not installed. Some visualization functions will not work.')


def pv_plot(meshes: list,
            image_2d=False,
            ve=None,
            cmap='viridis',
            plotter_kwargs: dict = None,
            add_mesh_kwargs: dict = None,
            background_plotter=False):
    """Function to plot meshes in vtk using pyvista

    Args:
        meshes (List[pv.PolyData]):
        image_2d (bool): If True convert plot to matplotlib imshow. This helps for visualizing
         the plot in IDEs
        ve (float): vertical exaggeration
        plotter_kwargs (dict): pyvista.Plotter kwargs
        add_mesh_kwargs (dict): pyvista.add_mesh kwargs
        background_plotter (bool): if true and pyvistaqt installed use pyvista
         backgroung plotter.
    """

    add_mesh_kwargs = dict() if add_mesh_kwargs is None else add_mesh_kwargs
    p: pv.Pll = init_plotter(image_2d, ve, plotter_kwargs)

    for m in meshes:
        # Check if m has texture data
        texture = None
        if hasattr(m, '_textures') and isinstance(m._textures, dict):
            texture = m._textures.get(0, None)
        
        p.add_mesh(
            mesh=m,
            cmap=cmap,
            categories=True,
            texture=texture,
            **add_mesh_kwargs
        )

    p.show_bounds()

    if image_2d is False:
        p.show()
        return p
    else:
        fig = pyvista_to_matplotlib(p)
        return fig


def pyvista_to_matplotlib(p: "pv.Plotter"):
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError('Matplotlib is necessary for generating a 2D image.')
    img = p.show(screenshot=True)
    img = p.last_image
    fig = plt.imshow(img)
    plt.axis('off')
    plt.show(block=False)
    p.close()
    return fig


def init_plotter(
        image_2d=False,
        ve=None,
        plotter_kwargs: dict = None
) -> "pv.Plotter":
    plotter_kwargs = dict() if plotter_kwargs is None else plotter_kwargs
    off_screen = True if image_2d is True else None
    p = pv.Plotter(**plotter_kwargs, off_screen=off_screen)
    if ve is not None:
        p.set_scale(zscale=ve)
    return p


def to_pyvista_points(point_set: PointSet):
    """Create pyvista.PolyData from PointSet

    Args:
        point_set (PointSet): Class for pointset based data structures.

    Returns:
        pv.PolyData
    """
    pv = optional_requirements.require_pyvista()
    poly: pv.PolyData = pv.PolyData(point_set.data.vertex)
    poly.point_data.update(point_set.data.attributes_to_dict)

    return poly


def to_pyvista_mesh(triangular_surface: TriSurf) -> "pv.PolyData":
    """Create planar surface PolyData from unstructured element such as TriSurf

    Returns:
        mesh texture
    """
    nve = triangular_surface.mesh.n_vertex_per_element
    vertices = triangular_surface.mesh.vertex

    # ? We need better name for these variables
    num_vertex_elements = np.full(triangular_surface.mesh.n_elements, nve)
    x = triangular_surface.mesh.cells

    cells = np.c_[num_vertex_elements, x]

    pv = optional_requirements.require_pyvista()
    mesh = pv.PolyData(vertices, cells)
    mesh.cell_data.update(triangular_surface.mesh.attributes_to_dict)
    mesh.point_data.update(triangular_surface.mesh.points_attributes)

    # If UV coordinates exist in points_attributes, set them as texture coordinates
    if triangular_surface.has_texture_data_with_uv:
        uv = np.column_stack((
            triangular_surface.mesh.points_attributes['u'],
            triangular_surface.mesh.points_attributes['v']
        ))
        mesh.active_texture_coordinates = uv
        if triangular_surface.texture is not None:
            texture_data = np.asarray(triangular_surface.texture.values, dtype=np.float32)
            mesh._textures = {0: texture_data}
            mesh.active_scalars_name = None

    elif triangular_surface.has_texture_data_without_uv:
        mesh.texture_map_to_plane(
            inplace=True,
            origin=triangular_surface.texture_origin,
            point_u=triangular_surface.texture_point_u,
            point_v=triangular_surface.texture_point_v
        )

        texture_data = np.asarray(triangular_surface.texture.values, dtype=np.float32)
        mesh._textures = {0: texture_data}
        mesh.active_scalars_name = None

    return mesh


def to_pyvista_mesh_and_texture(triangular_surface: Union[TriSurf], ) -> Tuple["pv.PolyData", Optional[np.array]]:
    """Create planar surface PolyData from unstructured element such as TriSurf

    Returns:
        mesh texture
    """
    mesh = to_pyvista_mesh(triangular_surface)

    if triangular_surface.texture is None:
        raise ValueError('unstructured_element needs texture data to be mapped.')

    mesh.texture_map_to_plane(
        inplace=True,
        origin=triangular_surface.texture_origin,
        point_u=triangular_surface.texture_point_u,
        point_v=triangular_surface.texture_point_v
    )
    tex = pv.numpy_to_texture(triangular_surface.texture.values)
    mesh._textures = {0: tex}

    from vtkmodules.util.numpy_support import vtk_to_numpy
    uv = vtk_to_numpy(mesh.GetPointData().GetTCoords())
    return mesh, uv


class PyvistaScalarType(enum.Enum):
    POINT = 'point'
    CELL = 'cell'


def to_pyvista_line(
        line_set: LineSet,
        as_tube=True,
        radius=None,
        spline=False,
        n_interp_points=1000,
        scalar_type: PyvistaScalarType = PyvistaScalarType.POINT,
        active_scalar: Optional[str] = None
):
    nve = line_set.data.n_vertex_per_element
    vertices = line_set.data.vertex
    cells = np.c_[np.full(line_set.data.n_elements, nve),
    line_set.data.cells]
    if spline is False:
        mesh = pv.PolyData()
        mesh.points = vertices
        mesh.lines = cells
    else:
        raise NotImplementedError

    match scalar_type:
        case PyvistaScalarType.POINT:
            mesh.point_data.update(line_set.data.points_attributes_to_dict)
            if active_scalar is not None:
                mesh.set_active_scalars(active_scalar, preference='point')
        case PyvistaScalarType.CELL:
            mesh.cell_data.update(line_set.data.attributes_to_dict)
            if active_scalar is not None:
                mesh.set_active_scalars(active_scalar, preference='cell')
    if as_tube is True:
        return mesh.tube(radius=radius)
    else:
        return mesh


def to_pyvista_tetra(tetra_mesh: TetraMesh):
    """Create pyvista.UnstructuredGrid"""
    vertices = tetra_mesh.data.vertex
    tets = tetra_mesh.data.cells
    cells = np.c_[np.full(len(tets), 4), tets]
    import vtk
    ctypes = np.array([vtk.VTK_TETRA, ], np.int32)
    mesh = pv.UnstructuredGrid(cells, ctypes, vertices)
    mesh.cell_data.update(tetra_mesh.data.attributes_to_dict)
    return mesh


def to_pyvista_grid(
        structured_grid: StructuredGrid,
        data_set_name: str = None,
        attribute_slice: dict = None,
        data_order: str = 'F'
) -> "pyvista.StructuredGrid":
    """

    Args:
        structured_grid:
        data_set_name:
        attribute_slice: dictionary to select which 3D array will be displayed as color

    Returns:

    """
    if attribute_slice is None:
        attribute_slice = dict()

    if data_set_name is None:
        data_set_name = structured_grid.ds.active_data_array_name

    cart_dims = structured_grid.cartesian_dimensions
    data_dims = structured_grid.ds.data[data_set_name].sel(**attribute_slice).ndim
    if cart_dims < data_dims:
        raise AttributeError('Data dimension and cartesian dimensions must match.'
                             'Possibly there are not valid dimension name in the'
                             'xarray.DataArray. These are X Y Z x y z')

    if data_dims == 2:
        meshgrid = structured_grid.meshgrid_2d(data_set_name)
    elif data_dims == 3:
        meshgrid = structured_grid.meshgrid_3d
    else:
        raise AttributeError('The DataArray does not have valid dimensionality. '
                             'Possibly there are not valid dimension name in the'
                             'xarray.DataArray. These are X Y Z x y z')

    pv = optional_requirements.require_pyvista()
    mesh = pv.StructuredGrid(*meshgrid)
    update_grid_attribute(mesh, structured_grid, data_order,
                          attribute_slice, data_set_name)

    return mesh


def update_grid_attribute(
        mesh: 'pv.StructuredGrid',
        structured_grid: StructuredGrid,
        data_order='F',
        attribute_slice=None,
        data_set_name=None
):
    if attribute_slice is None:
        attribute_slice = dict()

    if data_set_name is None:
        data_set_name = structured_grid.ds.active_data_array_name
    import xarray as xr
    dataset: xr.DataArray = structured_grid.ds.data[data_set_name]

    attributeData = {data_set_name: dataset.sel(**attribute_slice).values.ravel(data_order)}
    mesh.point_data.update(attributeData)

    return mesh


def _n_cartesian_coord(attribute, structured_grid):
    coord_names = np.array(['X', 'Y', 'Z', 'x', 'y', 'z'])
    ndim = np.isin(coord_names, structured_grid.ds.data[attribute].dims).sum()
    return ndim


def _generate_colors_from_colormap(num_colors, cmap_name='viridis'):
    """
    Generate a sequence of colors from a given Matplotlib colormap.

    Parameters:
    num_colors (int): Number of colors to generate.
    cmap_name (str): Name of the Matplotlib colormap to use.

    Returns:
    list of tuple: List of RGB color tuples.
    """
    import matplotlib.pyplot as plt
    colormap = plt.cm.get_cmap(cmap_name)
    colors = colormap(np.linspace(0, 1, num_colors))
    # Convert from RGBA to RGB and scale to 0-255
    return [(int(r * 255), int(g * 255), int(b * 255)) for r, g, b, _ in colors]



import numpy as np
import warnings

from subsurface.core.reader_helpers.readers_data import GenericReaderFilesHelper
import pandas as pd

from subsurface.modules.reader.wells._read_to_df import check_format_and_read_to_df
from subsurface.modules.reader.wells.wells_utils import add_tops_from_base_and_altitude_in_place


def read_collar(reader_helper: GenericReaderFilesHelper) -> pd.DataFrame:
    if reader_helper.index_col is False: reader_helper.index_col = 0

    # Check file_or_buffer type
    data_df: pd.DataFrame = check_format_and_read_to_df(reader_helper)
    _map_rows_and_cols_inplace(data_df, reader_helper)

    # Remove duplicates
    data_df = data_df[~data_df.index.duplicated(keep='first')]

    return data_df


def read_survey(reader_helper: GenericReaderFilesHelper, validate_survey: bool = True) -> pd.DataFrame:
    if reader_helper.index_col is False: reader_helper.index_col = 0

    d = check_format_and_read_to_df(reader_helper)
    _map_rows_and_cols_inplace(d, reader_helper)

    if validate_survey:
        d_no_singles = _validate_survey_data(d)
    else:
        d_no_singles = d

    return d_no_singles


def read_lith(reader_helper: GenericReaderFilesHelper) -> pd.DataFrame:
    return read_attributes(reader_helper, is_lith=True)


def read_attributes(reader_helper: GenericReaderFilesHelper, is_lith: bool = False, validate_attr: bool = True) -> pd.DataFrame:
    if reader_helper.index_col is False:
        reader_helper.index_col = 0
        
    d = check_format_and_read_to_df(reader_helper)

    _map_rows_and_cols_inplace(d, reader_helper)
    if validate_attr is False:
        return d
    
    if is_lith:
        d = _validate_lith_data(d, reader_helper)
    else:
        _validate_attr_data(d)
    return d


def _map_rows_and_cols_inplace(d: pd.DataFrame, reader_helper: GenericReaderFilesHelper):
    if reader_helper.index_map is not None:
        d.rename(reader_helper.index_map, axis="index", inplace=True)  # d.index = d.index.map(reader_helper.index_map)
    if reader_helper.columns_map is not None:
        d.rename(reader_helper.columns_map, axis="columns", inplace=True)


def _validate_survey_data(d):
    # Check for essential column 'md'
    if 'md' not in d.columns:
        raise AttributeError(
            'md, inc, and azi columns must be present in the file. Use columns_map to assign column names to these fields.')

    # Check if 'dip' column exists and convert it to 'inc'
    if 'dip' in d.columns:
        # Convert dip to inclination (90 - dip)
        d['inc'] = 90 - d['dip']
        # Optionally, drop the 'dip' column if it's no longer needed
        d.drop(columns=['dip'], inplace=True)

    # Handle if inclination ('inc') or azimuth ('azi') columns are missing
    if not np.isin(['inc', 'azi'], d.columns).all():
        warnings.warn(
            'inc and/or azi columns are not present in the file. The boreholes will be straight.')
        d['inc'] = 180
        d['azi'] = 0

    # Drop wells that contain only one value, ensuring that we keep rows only when there are duplicates
    d_no_singles = d[d.index.duplicated(keep=False)]

    return d_no_singles


def _validate_attr_data(d):
    assert d.columns.isin(['base']).any(), ('base column must be present in the file. '
                                            'Use columns_map to assign column names to these fields.')


def _validate_lith_data(d: pd.DataFrame, reader_helper: GenericReaderFilesHelper) -> pd.DataFrame:
    # Check component lith in column
    if 'component lith' not in d.columns:
        raise AttributeError('If wells attributes represent lithology, `component lith` column must be present in the file. '
                             'Use columns_map to assign column names to these fields. Maybe you are marking as lithology'
                             'the wrong file?')
    else:
        # TODO: Add categories to reader helper
        categories = sorted(d['component lith'].dropna().unique())
        d['component lith'] = pd.Categorical(
            d['component lith'],
            categories=categories,
            ordered=True
        )
        
        d['lith_ids'] = d['component lith'].cat.codes + 1

    given_top = np.isin(['top', 'base'], d.columns).all()
    given_altitude_and_base = np.isin(['altitude', 'base'], d.columns).all()
    given_only_base = np.isin(['base'], d.columns).all()
    if given_altitude_and_base and not given_top:
        warnings.warn('top column is not present in the file. The tops will be calculated from the base and altitude')
        d = add_tops_from_base_and_altitude_in_place(
            data=d,
            col_well_name=reader_helper.index_col,
            col_base='base',
            col_altitude='altitude'
        )
    elif given_only_base and not given_top:
        warnings.warn('top column is not present in the file. The tops will be calculated from the base assuming altitude=0')
        # add a top column with 0 and call add_tops_from_base_and_altitude_in_place
        d['altitude'] = 0
        d = add_tops_from_base_and_altitude_in_place(
            data=d,
            col_well_name=reader_helper.index_col,
            col_base='base',
            col_altitude='altitude'
        )


    elif not given_top and not given_altitude_and_base:
        raise ValueError('top column or base and altitude columns must be present in the file. '
                         'Use columns_map to assign column names to these fields. Maybe you are marking as lithology'
                         'the wrong file?')

    lith_df = d[['top', 'base', 'component lith']]

    # * Make sure values are positive
    lith_df['top'] = np.abs(lith_df['top'])
    lith_df['base'] = np.abs(lith_df['base'])

    return lith_df



from typing import List

import pandas as pd


__all__ = ['add_tops_from_base_and_altitude_in_place',
           'fix_wells_higher_base_than_top_inplace', 'map_attr_to_segments',
           'pivot_wells_df_into_segment_per_row']


def add_tops_from_base_and_altitude_in_place(data: pd.DataFrame, col_well_name: str, col_base: str,
                                             col_altitude: str) -> pd.DataFrame:
    d = data
    d = _remove_repeated_rows(d)
    _create_base_col(col_altitude, col_base, d)
    _create_top_col(col_altitude, col_well_name, d)
    _add_md_col_if_missing(d)
    return d


def fix_wells_higher_base_than_top_inplace(df_fixed) -> pd.DataFrame:
    top_base_error = df_fixed["top"] > df_fixed["base"]
    df_fixed["base"][top_base_error] = df_fixed["top"] + 0.01
    return df_fixed


def map_attr_to_segments(df, attr_per_segment: List, n_wells: int) -> pd.DataFrame:
    tiled_formations = pd.np.tile(attr_per_segment, (n_wells))
    df['formation'] = tiled_formations
    return df


def pivot_wells_df_into_segment_per_row(df: pd.DataFrame, start_segment_cols: int, n_segments_per_well: int) -> pd.DataFrame:
    # Repeat fixed rows (collar name and so)
    df_fixed = df.iloc[:, :start_segment_cols]
    df_fixed = df_fixed.loc[df_fixed.index.repeat(n_segments_per_well)]

    df_bottoms = df.iloc[:, start_segment_cols:start_segment_cols + n_segments_per_well]
    df_fixed['base'] = df_bottoms.values.reshape(-1, 1, order='C')

    return df_fixed


def _add_md_col_if_missing(d):
    if "md" not in d.columns:
        d.loc[:, 'md'] = d['top']


def _create_top_col(col_altitude, col_well_name, d):
    Z_shift = d.groupby(col_well_name)['base'].shift(1)
    Z_0 = Z_shift.fillna(0)
    v = Z_0 + d[col_altitude]
    d.loc[:, 'top'] = Z_0
    d.loc[:, '_top_abs'] = v


def _create_base_col(col_altitude, col_base, d):
    d.loc[:, 'base'] = d[col_altitude] - d[col_base]


def _remove_repeated_rows(d):
    repeated_rows = _mark_repeated_rows(d['base'])
    d = d[~repeated_rows]  # removed repeated rows
    return d


def _mark_repeated_rows(series: pd.Series):
    return series.shift(1) == series



import enum
import numpy as np
import pandas as pd

from .collars import Collars
from .survey import Survey
from ...structs import LineSet
from ...structs.base_structures import UnstructuredData


class MergeOptions(enum.Enum):
    RAISE = enum.auto()
    INTERSECT = enum.auto()


def create_combined_trajectory(collars: Collars, survey: Survey, merge_option: MergeOptions, slice_: slice):
    collar_df = _create_collar_df(collars, slice_, survey.well_id_mapper)
    survey_df_vertex = _create_survey_df(survey)

    if merge_option == MergeOptions.RAISE:
        raise NotImplementedError("RAISE merge option not implemented")
        _validate_matching_ids(collar_df, survey_df_vertex)
    elif merge_option == MergeOptions.INTERSECT:
        return _Intersect.process_intersection(collar_df, survey_df_vertex, survey)
    else:
        raise ValueError("Unsupported merge option")


def _create_collar_df(collars, slice_, well_id_mapper: dict ):
    collar_df = pd.DataFrame(collars.collar_loc.points[slice_], columns=['X', 'Y', 'Z'])
    selected_collars:list = collars.ids[slice_]
    collar_df['well_id'] = pd.Series(selected_collars).map(well_id_mapper)
    return collar_df


def _create_survey_df(survey):
    survey_df_vertex = pd.DataFrame(survey.survey_trajectory.data.vertex, columns=['X', 'Y', 'Z'])
    vertex_attrs = survey.survey_trajectory.data.points_attributes
    id_int_vertex = vertex_attrs['well_id']
    survey_df_vertex['well_id'] = id_int_vertex.map(pd.Series(survey.ids))
    return survey_df_vertex


def _validate_matching_ids(collar_df, survey_df_vertex):
    missing_from_survey = set(collar_df['well_id']) - set(survey_df_vertex['well_id'])
    missing_from_collar = set(survey_df_vertex['well_id']) - set(collar_df['well_id'])
    if missing_from_survey or missing_from_collar:
        raise ValueError(f"Collars and survey ids do not match. Missing in survey: {missing_from_survey}, Missing in collars: {missing_from_collar}")


class _Intersect:
    """This class is just to create a namespace for the intersection method"""
    @staticmethod
    def process_intersection(collar_df: pd.DataFrame, survey_df_vertex: pd.DataFrame, survey: Survey) -> LineSet:
        
        combined_df_vertex = pd.merge(
            left=survey_df_vertex,
            right=collar_df,
            how='outer',
            indicator=True,
            on='well_id',
            suffixes=('_collar', '_survey')
        )
        combined_df_vertex = combined_df_vertex[combined_df_vertex['_merge'].isin(['left_only', 'both']) ]

        vertex_attrs = survey.survey_trajectory.data.points_attributes
        if len(combined_df_vertex) != len(vertex_attrs):
            raise ValueError("Vertex and vertex attributes have different lengths")
        _Intersect._add_collar_coordinates(combined_df_vertex)

        combined_df_cells = _Intersect._generate_cells(combined_df_vertex, survey)
        cell_attributes = survey.survey_trajectory.data.cell_attributes
        if len(combined_df_cells) != len(cell_attributes):
            raise ValueError("Cells and cell attributes have different lengths")

        line_set: LineSet = _Intersect._create_line_set(combined_df_vertex, combined_df_cells, survey)

        line_set.data.data.attrs.update(survey.survey_trajectory.data.data.attrs)
        line_set.data.data.attrs.update(collar_df.attrs)
        
        return line_set

    @staticmethod
    def _add_collar_coordinates(combined_df_vertex: pd.DataFrame):
        combined_df_vertex['X_survey'] += combined_df_vertex['X_collar']
        combined_df_vertex['Y_survey'] += combined_df_vertex['Y_collar']
        combined_df_vertex['Z_survey'] += combined_df_vertex['Z_collar']

    @staticmethod
    def _generate_cells(combined_df_vertex: pd.DataFrame, survey: Survey) -> pd.DataFrame:
        combined_df_cells = []
        previous_index = 0
        for e, well_id in enumerate(survey.ids):
            df_vertex_well = combined_df_vertex[combined_df_vertex['well_id'] == well_id]
            indices = np.arange(len(df_vertex_well)) + previous_index
            previous_index += len(df_vertex_well)
            cells = np.array([indices[:-1], indices[1:]]).T
            df_cells_well = pd.DataFrame(cells, columns=['cell1', 'cell2'])
            df_cells_well['well_id'] = well_id
            df_cells_well['well_id_int'] = e
            combined_df_cells.append(df_cells_well)

        return pd.concat(combined_df_cells, ignore_index=True)

    @staticmethod
    def _create_line_set(combined_df_vertex: pd.DataFrame, combined_df_cells: pd.DataFrame, survey: Survey) -> LineSet:
        vertex_attrs = survey.survey_trajectory.data.points_attributes
        cell_attributes = survey.survey_trajectory.data.cell_attributes

        combined_trajectory_unstruct = UnstructuredData.from_array(
            vertex=combined_df_vertex[['X_survey', 'Y_survey', 'Z_survey']].values,
            cells=combined_df_cells[['cell1', 'cell2']].values,
            vertex_attr=vertex_attrs,
            cells_attr=cell_attributes
        )

        return LineSet(data=combined_trajectory_unstruct, radius=500)



import warnings

from typing import Union, Hashable, Optional

import pandas as pd
from dataclasses import dataclass
import numpy as np
import xarray as xr

from subsurface import optional_requirements
from ...structs.unstructured_elements import LineSet
from ...structs.base_structures import UnstructuredData

NUMBER_NODES = 30
RADIUS = 10


@dataclass
class Survey:
    ids: list[str]
    survey_trajectory: LineSet
    well_id_mapper: dict[str, int] = None  #: This is following the order of the survey csv that can be different that the collars

    @property
    def id_to_well_id(self):
        # Reverse the well_id_mapper dictionary to map IDs to well names
        id_to_well_name_mapper = {v: k for k, v in self.well_id_mapper.items()}
        return id_to_well_name_mapper

    @classmethod
    def from_df(cls, survey_df: 'pd.DataFrame', attr_df: Optional['pd.DataFrame'] = None, number_nodes: Optional[int] = NUMBER_NODES,
                duplicate_attr_depths: bool = False) -> 'Survey':
        """
        Create a Survey object from two DataFrames containing survey and attribute data.

        :param survey_df: DataFrame containing survey data.
        :param attr_df: DataFrame containing attribute data. This is used to make sure the raw data is perfectly aligned.
        :param number_nodes: Optional parameter specifying the number of nodes.
        :return: A Survey object representing the input data.

        """
        trajectories: UnstructuredData = _data_frame_to_unstructured_data(
            survey_df=_correct_angles(survey_df),
            attr_df=attr_df,
            number_nodes=number_nodes,
            duplicate_attr_depths=duplicate_attr_depths
        )
        # Grab the unique ids
        unique_ids = trajectories.points_attributes["well_id"].unique()

        return cls(
            ids=unique_ids,
            survey_trajectory=LineSet(data=trajectories, radius=RADIUS),
            well_id_mapper=trajectories.data.attrs["well_id_mapper"]
        )

    def get_well_string_id(self, well_id: int) -> str:
        return self.ids[well_id]

    def get_well_num_id(self, well_string_id: Union[str, Hashable]) -> int:
        return self.well_id_mapper.get(well_string_id, None)

    def update_survey_with_lith(self, lith: pd.DataFrame):
        unstruct: UnstructuredData = _combine_survey_and_attrs(lith, self)
        self.survey_trajectory.data = unstruct

    def update_survey_with_attr(self, attrs: pd.DataFrame):
        self.survey_trajectory.data = _combine_survey_and_attrs(attrs, self)


def _combine_survey_and_attr(lith: pd.DataFrame, survey: Survey) -> UnstructuredData:
    pass


def _combine_survey_and_attrs(attrs: pd.DataFrame, survey: Survey) -> UnstructuredData:
    # Import moved to top for clarity and possibly avoiding repeated imports if called multiple times
    from ...structs.base_structures._unstructured_data_constructor import raw_attributes_to_dict_data_arrays

    # Accessing trajectory data more succinctly
    trajectory: xr.DataArray = survey.survey_trajectory.data.data["vertex_attrs"]
    # Ensure all columns in lith exist in new_attrs, if not, add them as NaN

    new_attrs = _map_attrs_to_measured_depths(attrs, survey)

    # Construct the final xarray dict without intermediate variable
    points_attributes_xarray_dict = raw_attributes_to_dict_data_arrays(
        default_attributes_name="vertex_attrs",
        n_items=trajectory.shape[0],  # TODO: Can I look this on new_attrs to remove line 11?
        dims=["points", "vertex_attr"],
        raw_attributes=new_attrs
    )

    # Inline construction of UnstructuredData
    return UnstructuredData.from_data_arrays_dict(
        xarray_dict={
                "vertex"      : survey.survey_trajectory.data.data["vertex"],
                "cells"       : survey.survey_trajectory.data.data["cells"],
                "vertex_attrs": points_attributes_xarray_dict["vertex_attrs"],
                "cell_attrs"  : survey.survey_trajectory.data.data["cell_attrs"]
        },
        xarray_attributes=survey.survey_trajectory.data.data.attrs,
        default_cells_attributes_name=survey.survey_trajectory.data.cells_attr_name,
        default_points_attributes_name=survey.survey_trajectory.data.vertex_attr_name
    )


def _map_attrs_to_measured_depths(attrs: pd.DataFrame, survey: Survey) -> pd.DataFrame:
    trajectory: xr.DataArray = survey.survey_trajectory.data.data["vertex_attrs"]
    trajectory_well_id: xr.DataArray = trajectory.sel({'vertex_attr': 'well_id'})
    measured_depths: np.ndarray = trajectory.sel({'vertex_attr': 'measured_depths'}).values.astype(np.float64)

    # Start with a copy of the existing attributes DataFrame
    new_attrs = survey.survey_trajectory.data.points_attributes.copy()
    if 'component lith' in attrs.columns and 'lith_ids' not in attrs.columns:
        # Factorize lith components directly in-place
        attrs['lith_ids'], _ = pd.factorize(attrs['component lith'], use_na_sentinel=True)
    else:
        pass

    # Add missing columns from attrs, preserving their dtypes
    for col in attrs.columns.difference(new_attrs.columns):
        new_attrs[col] = np.nan if pd.api.types.is_numeric_dtype(attrs[col]) else None

    # Align well IDs between attrs and trajectory, perform interpolation, and map the attributes
    # Loop dict
    for survey_well_name in survey.well_id_mapper:
        # Select rows corresponding to the current well ID

        # use the well_id to get all the elements of attrs that have the well_id as index
        if survey_well_name in attrs.index:
            attrs_well = attrs.loc[[survey_well_name]]
            # Proceed with processing attrs_well
        else:
            print(f"Well '{survey_well_name}' does not exist in the attributes DataFrame.")
            continue

        survey_well_id = survey.get_well_num_id(survey_well_name)
        trajectory_well_mask = (trajectory_well_id == survey_well_id).values

        # Apply mask to measured depths for the current well
        well_measured_depths = measured_depths[trajectory_well_mask]

        if "base" not in attrs_well.columns:
            raise ValueError(f"Base column must be present in the file for well '{survey_well_name}'.")
        elif "top" not in attrs_well.columns:
            location_values_to_interpolate = attrs_well['base']
        else:
            location_values_to_interpolate = (attrs_well['top'] + attrs_well['base']) / 2

        # Interpolation for each attribute column
        for col in attrs_well.columns:
            # Interpolate the attribute values based on the measured depths
            if col in ['top', 'base', 'well_id']:
                continue
            attr_to_interpolate = attrs_well[col]
            # make sure the attr_to_interpolate is not a string
            if attr_to_interpolate.dtype == 'O' or isinstance(attr_to_interpolate.dtype, pd.CategoricalDtype): 
                continue
            if col in ['lith_ids', 'component lith']:
                interp_kind = 'nearest'
            else:
                interp_kind = 'linear'

            from scipy.interpolate import interp1d
            interp_func = interp1d(
                x=location_values_to_interpolate,
                y=attr_to_interpolate,
                bounds_error=False,
                fill_value=np.nan,
                kind=interp_kind
            )

            # Assign the interpolated values to the new_attrs DataFrame
            vals = interp_func(well_measured_depths)
            new_attrs.loc[trajectory_well_mask, col] = vals

    return new_attrs


def _map_attrs_to_measured_depths_(attrs: pd.DataFrame, new_attrs: pd.DataFrame, survey: Survey):
    warnings.warn("This function is obsolete. Use _map_attrs_to_measured_depths instead.", DeprecationWarning)

    trajectory: xr.DataArray = survey.survey_trajectory.data.data["vertex_attrs"]
    well_ids: xr.DataArray = trajectory.sel({'vertex_attr': 'well_id'})
    measured_depths: xr.DataArray = trajectory.sel({'vertex_attr': 'measured_depths'})

    new_columns = attrs.columns.difference(new_attrs.columns)
    new_attrs = pd.concat([new_attrs, pd.DataFrame(columns=new_columns)], axis=1)
    for index, row in attrs.iterrows():
        well_id = survey.get_well_num_id(index)
        if well_id is None:
            print(f'Well ID {index} not found in survey trajectory. Skipping lithology assignment.')

        well_id_mask = well_ids == well_id

        # TODO: Here we are going to need to interpolate

        spatial_mask = ((measured_depths <= row['top']) & (measured_depths >= row['base']))
        mask = well_id_mask & spatial_mask

        new_attrs.loc[mask.values, attrs.columns] = row.values
    return new_attrs


def _correct_angles(df: pd.DataFrame) -> pd.DataFrame:
    def correct_inclination(inc: float) -> float:
        if inc < 0:
            inc = inc % 360  # Normalize to 0-360 range first if negative
        if 0 <= inc <= 180:
            # add or subtract a very small number to make sure that 0 or 180 are never possible
            return inc + 1e-10 if inc == 0 else inc - 1e-10
        elif 180 < inc < 360:
            return 360 - inc  # Reflect angles greater than 180 back into the 0-180 range
        else:
            raise ValueError(f'Inclination value {inc} is out of the expected range of 0 to 360 degrees')

    def correct_azimuth(azi: float) -> float:
        return azi % 360  # Normalize azimuth to 0-360 range

    df['inc'] = df['inc'].apply(correct_inclination)
    df['azi'] = df['azi'].apply(correct_azimuth)

    return df


def _data_frame_to_unstructured_data(survey_df: 'pd.DataFrame', number_nodes: int, attr_df: Optional['pd.DataFrame'] = None,
                                     duplicate_attr_depths: bool = False) -> UnstructuredData:

    wp = optional_requirements.require_wellpathpy()

    cum_vertex: np.ndarray = np.empty((0, 3), dtype=np.float32)
    cells: np.ndarray = np.empty((0, 2), dtype=np.int_)
    cell_attr: pd.DataFrame = pd.DataFrame(columns=['well_id'], dtype=np.float32)
    vertex_attr: pd.DataFrame = pd.DataFrame()

    for e, (borehole_id, data) in enumerate(survey_df.groupby(level=0)):
        dev = wp.deviation(
            md=data['md'].values,
            inc=data['inc'].values,
            azi=data['azi'].values
        )

        md_min = dev.md.min()
        md_max = dev.md.max()

        attr_depths = _grab_depths_from_attr(
            attr_df=attr_df,
            borehole_id=borehole_id,
            duplicate_attr_depths=duplicate_attr_depths,
            md_max=md_max,
            md_min=md_min
        )

        # Now combine attr_depths with depths
        md_min = dev.md.min()
        md_max = dev.md.max()
        depths = np.linspace(md_min, md_max, number_nodes)
        depths = np.union1d(depths, attr_depths)
        depths.sort()

        # Resample positions at depths
        pos = dev.minimum_curvature().resample(depths=depths)
        vertex_count = cum_vertex.shape[0]

        this_well_vertex = np.vstack([pos.easting, pos.northing, pos.depth]).T
        cum_vertex = np.vstack([cum_vertex, this_well_vertex])
        measured_depths = _calculate_distances(array_of_vertices=this_well_vertex)

        n_vertex_shift_0 = np.arange(0, len(pos.depth) - 1, dtype=np.int_)
        n_vertex_shift_1 = np.arange(1, len(pos.depth), dtype=np.int_)
        cell_per_well = np.vstack([n_vertex_shift_0, n_vertex_shift_1]).T + vertex_count
        cells = np.vstack([cells, cell_per_well])

        attribute_values = np.isin(depths, attr_depths)

        vertex_attr_per_well = pd.DataFrame({
                'well_id'        : [e] * len(pos.depth),
                'measured_depths': measured_depths,
                'is_attr_point'  : attribute_values,
        })

        vertex_attr = pd.concat([vertex_attr, vertex_attr_per_well], ignore_index=True)

        # Add the id (e), to cell_attr
        cell_attr = pd.concat([cell_attr, pd.DataFrame({'well_id': [e] * len(cell_per_well)})], ignore_index=True)

    unstruct = UnstructuredData.from_array(
        vertex=cum_vertex,
        cells=cells.astype(int),
        vertex_attr=vertex_attr.reset_index(drop=True),
        cells_attr=cell_attr.reset_index(drop=True)
    )

    unstruct.data.attrs["well_id_mapper"] = {well_id: e for e, well_id in enumerate(survey_df.index.unique(level=0))}

    return unstruct


def _grab_depths_from_attr(
        attr_df: pd.DataFrame,
        borehole_id: Hashable,
        duplicate_attr_depths: bool,
        md_max: float,
        md_min: float
) -> np.ndarray:
    # Initialize attr_depths and attr_labels as empty arrays
    attr_depths = np.array([], dtype=float)
    attr_labels = np.array([], dtype='<U4')  # Initialize labels for 'top' and 'base'

    if attr_df is None or ("top" not in attr_df.columns and "base" not in attr_df.columns):
        return attr_depths

    try:
        vals = attr_df.loc[borehole_id]

        tops = np.array([], dtype=float)
        bases = np.array([], dtype=float)

        if 'top' in vals:
            if isinstance(vals, pd.DataFrame):
                tops = vals['top'].values.flatten()
            else:
                tops = np.array([vals['top']])
            # Convert to float and remove NaNs
            tops = tops.astype(float)
            tops = tops[~np.isnan(tops)]
            # Clip to within md range
            tops = tops[(tops >= md_min) & (tops <= md_max)]

        if 'base' in vals:
            if isinstance(vals, pd.DataFrame):
                bases = vals['base'].values.flatten()
            else:
                bases = np.array([vals['base']])
            # Convert to float and remove NaNs
            bases = bases.astype(float)
            bases = bases[~np.isnan(bases)]
            # Clip to within md range
            bases = bases[(bases >= md_min) & (bases <= md_max)]

        # Combine tops and bases into attr_depths with labels
        attr_depths = np.concatenate((tops, bases))
        attr_labels = np.array(['top'] * len(tops) + ['base'] * len(bases))

        # Drop duplicates while preserving order
        _, unique_indices = np.unique(attr_depths, return_index=True)
        attr_depths = attr_depths[unique_indices]
        attr_labels = attr_labels[unique_indices]

    except KeyError:
        # No attributes for this borehole_id or missing columns
        attr_depths = np.array([], dtype=float)
        attr_labels = np.array([], dtype='<U4')

    # If duplicate_attr_depths is True, duplicate attr_depths with a tiny offset
    if duplicate_attr_depths and len(attr_depths) > 0:
        tiny_offset = (md_max - md_min) * 1e-6  # A tiny fraction of the depth range
        # Create offsets: +tiny_offset for 'top', -tiny_offset for 'base'
        offsets = np.where(attr_labels == 'top', tiny_offset, -tiny_offset)
        duplicated_attr_depths = attr_depths + offsets
        # Ensure the duplicated depths are within the md range
        valid_indices = (duplicated_attr_depths >= md_min) & (duplicated_attr_depths <= md_max)
        duplicated_attr_depths = duplicated_attr_depths[valid_indices]
        # Original attribute depths
        original_attr_depths = attr_depths
        # Combine originals and duplicates
        attr_depths = np.hstack([original_attr_depths, duplicated_attr_depths])

    return attr_depths


def _calculate_distances(array_of_vertices: np.ndarray) -> np.ndarray:
    # Calculate the differences between consecutive points
    differences = np.diff(array_of_vertices, axis=0)

    # Calculate the Euclidean distance for each pair of consecutive points
    distances = np.linalg.norm(differences, axis=1)
    # Insert a 0 at the beginning to represent the starting point at the surface
    measured_depths = np.insert(np.cumsum(distances), 0, 0)
    return measured_depths



from ..base_structures import UnstructuredData


class TetraMesh:
    """PointSet with tetrahedron cells.

    This dataset defines cell connectivity between points to create
    tetrahedrons. This is volumetric.

    Args:

        data (UnstructuredData): Base object for unstructured data.

         data.cells represent the indices of the points for each
         tetrahedron in the mesh. Each column corresponds to a tetrahedron.
         Every tetrahedron is defined by the four points; where the first
         three (0,1,2) are the base of the tetrahedron which, using the
         right hand rule, forms a triangle whose normal points in the
         direction of the fourth point.

    """

    def __init__(self, data: UnstructuredData):
        if data.cells.shape[1] != 4:
            raise AttributeError('data.cells must be of the format'
                                 'NDArray[(Any, 4), IntX]')
        self.data = data

    @property
    def tetrahedrals(self):
        return self.data.cells

    @property
    def n_tetrahedrals(self):
        return self.tetrahedrals.shape[0]



from typing import List

from dataclasses import dataclass

from subsurface.core.reader_helpers.readers_data import GenericReaderFilesHelper


@dataclass
class ReaderWellsHelper:
    reader_collars_args: GenericReaderFilesHelper
    reader_survey_args: GenericReaderFilesHelper
    reader_lith_args: GenericReaderFilesHelper = None
    reader_attr_args: List[GenericReaderFilesHelper] = None



import json

import numpy as np
import xarray as xr


def read_data(json_file_path, binary_file_path):
    header = _read_json_header(json_file_path)
    data_dict = _read_binary_data(binary_file_path, header)
    return data_dict, header


def _read_json_header(json_file_path):
    with open(json_file_path, 'r') as f:
        header = json.load(f)
    return header


def _read_binary_data(binary_file_path, header, order='F'):
    # Initialize offset to 0
    offset = 0
    # Create a dictionary to hold the arrays
    data_dict = {}

    # Loop over the arrays
    for array_name in ['vertex', 'cell', 'cell_attr', 'vertex_attr']:
        # Get the shape and type of the array from the header
        array_shape = header.get(f'{array_name}_shape')
        array_type = header.get(f'{array_name}_types', 'float32')  # default to float32 if not found

        # Calculate the number of elements in the array
        num_elements = np.prod(array_shape)

        # Read the data
        data = np.fromfile(binary_file_path, dtype="int32", count=num_elements, offset=offset)
        data = data.reshape(array_shape, order=order)

        # Store the data in the dictionary
        data_dict[array_name] = data

        # Update the offset
        offset += data.nbytes

    return data_dict


